@@ -45,6 +45,7 @@
 #include "tcp_options.h"
 #include "core_cmd.h"
 #include "cfg_core.h"
+#include "ppcfg.h"
 
 #ifdef USE_DNS_CACHE
 void dns_cache_debug(rpc_t* rpc, void* ctx);
@@ -72,12 +73,12 @@ static const char* dns_cache_mem_info_doc[] = {
 	0                      /* Method signature(s) */
 };
 static const char* dns_cache_debug_doc[] = {
-	"dns debug  info.",    /* Documentation string */
+	"dns debug info.",    /* Documentation string */
 	0                      /* Method signature(s) */
 };
 
 static const char* dns_cache_debug_all_doc[] = {
-	"complete dns debug  dump",    /* Documentation string */
+	"complete dns debug dump",    /* Documentation string */
 	0                              /* Method signature(s) */
 };
 
@@ -184,7 +185,7 @@ static const char* dns_get_server_state_doc[] = {
 
 #endif /* DNS_WATCHDOG_SUPPORT */
 #endif /* USE_DNS_CACHE */
-#ifdef USE_DST_BLACKLIST
+#ifdef USE_DST_BLOCKLIST
 void dst_blst_debug(rpc_t* rpc, void* ctx);
 void dst_blst_mem_info(rpc_t* rpc, void* ctx);
 void dst_blst_view(rpc_t* rpc, void* ctx);
@@ -192,39 +193,38 @@ void dst_blst_delete_all(rpc_t* rpc, void* ctx);
 void dst_blst_add(rpc_t* rpc, void* ctx);
 
 static const char* dst_blst_mem_info_doc[] = {
-	"dst blacklist memory usage info.",  /* Documentation string */
+	"dst blocklist memory usage info.",  /* Documentation string */
 	0                                    /* Method signature(s) */
 };
 static const char* dst_blst_debug_doc[] = {
-	"dst blacklist  debug  info.",  /* Documentation string */
+	"dst blocklist debug info.",  /* Documentation string */
 	0                               /* Method signature(s) */
 };
 static const char* dst_blst_view_doc[] = {
-	"dst blacklist dump in human-readable format.",  /* Documentation string */
+	"dst blocklist dump in human-readable format.",  /* Documentation string */
 	0                               /* Method signature(s) */
 };
 static const char* dst_blst_delete_all_doc[] = {
-	"Deletes all the entries from the dst blacklist except the permanent ones.",  /* Documentation string */
+	"Deletes all the entries from the dst blocklist except the permanent ones.",  /* Documentation string */
 	0                               /* Method signature(s) */
 };
 static const char* dst_blst_add_doc[] = {
-	"Adds a new entry to the dst blacklist.",  /* Documentation string */
+	"Adds a new entry to the dst blocklist.",  /* Documentation string */
 	0                               /* Method signature(s) */
 };
-#ifdef USE_DST_BLACKLIST_STATS
+#ifdef USE_DST_BLOCKLIST_STATS
 void dst_blst_stats_get(rpc_t* rpc, void* ctx);
 
 static const char* dst_blst_stats_get_doc[] = {
-	"returns the dst blacklist measurement counters.",
+	"returns the dst blocklist measurement counters.",
 	0
 };
-#endif /* USE_DST_BLACKLIST_STATS */
+#endif /* USE_DST_BLOCKLIST_STATS */
 
 #endif
 
 
-
-#define MAX_CTIME_LEN 128
+#define MAX_CTIME_LEN 26
 
 /* up time */
 static char up_since_ctime[MAX_CTIME_LEN];
@@ -269,7 +269,7 @@ static void system_methodHelp(rpc_t* rpc, void* c)
 		rpc->fault(c, 400, "Method Name Expected");
 		return;
 	}
-	
+
 	r=rpc_lookup(name, strlen(name));
 	if (r==0){
 		rpc->fault(c, 400, "command not found");
@@ -380,13 +380,14 @@ static void core_uptime(rpc_t* rpc, void* c)
 {
 	void* s;
 	time_t now;
+	char buf[MAX_CTIME_LEN];
 	str snow;
+	snow.s = buf;
 
 	time(&now);
 
 	if (rpc->add(c, "{", &s) < 0) return;
-	snow.s = ctime(&now);
-	if(snow.s) {
+	if(ctime_r(&now, snow.s)) {
 		snow.len = strlen(snow.s);
 		if(snow.len>2 && snow.s[snow.len-1]=='\n') snow.len--;
 		rpc->struct_add(s, "S", "now", &snow);
@@ -437,6 +438,28 @@ static void core_psx(rpc_t* rpc, void* c)
 	}
 }
 
+static const char* core_psa_doc[] = {
+	"Return all the attributes of running.",
+		/* Documentation string */
+	0	/* Method signature(s) */
+};
+
+
+static void core_psa(rpc_t* rpc, void* c)
+{
+	int p;
+	void *handle;
+
+	for (p=0; p<*process_count;p++) {
+		rpc->add(c, "{", &handle);
+		rpc->struct_add(handle, "dddds",
+				"index", p,
+				"pid", pt[p].pid,
+				"status", pt[p].status,
+				"rank", pt[p].rank,
+				"description", pt[p].desc);
+	}
+}
 
 static const char* core_pwd_doc[] = {
 	"Returns the working directory of server.",    /* Documentation string */
@@ -446,23 +469,23 @@ static const char* core_pwd_doc[] = {
 
 static void core_pwd(rpc_t* rpc, void* c)
 {
-        char *cwd_buf;
-        int max_len;
+	char *cwd_buf;
+	int max_len;
 
-        max_len = pathmax();
-        cwd_buf = pkg_malloc(max_len);
-        if (!cwd_buf) {
-                ERR("core_pwd: No memory left\n");
-                rpc->fault(c, 500, "Server Ran Out of Memory");
+	max_len = pathmax();
+	cwd_buf = pkg_malloc(max_len);
+	if (!cwd_buf) {
+		PKG_MEM_ERROR;
+		rpc->fault(c, 500, "Server Ran Out of Memory");
 		return;
-        }
+	}
 
-        if (getcwd(cwd_buf, max_len)) {
+	if (getcwd(cwd_buf, max_len)) {
 		rpc->add(c, "s", cwd_buf);
-        } else {
+	} else {
 		rpc->fault(c, 500, "getcwd Failed");
-        }
-        pkg_free(cwd_buf);
+	}
+	pkg_free(cwd_buf);
 }
 
 
@@ -474,11 +497,11 @@ static const char* core_arg_doc[] = {
 
 static void core_arg(rpc_t* rpc, void* c)
 {
-        int p;
+	int p;
 
-        for (p = 0; p < my_argc; p++) {
+	for (p = 0; p < my_argc; p++) {
 		if (rpc->add(c, "s", my_argv[p]) < 0) return;
-        }
+	}
 }
 
 
@@ -658,7 +681,7 @@ all:
 
 
 static const char* core_sfmalloc_doc[] = {
-	"Returns sfmalloc debugging  info.",  /* Documentation string */
+	"Returns sfmalloc debugging info.",  /* Documentation string */
 	0                                     /* Method signature(s) */
 };
 
@@ -876,14 +899,16 @@ static const char* core_aliases_list_doc[] = {
 static void core_aliases_list(rpc_t* rpc, void* c)
 {
 	void *hr;
+	void *hs;
 	void *ha;
 	struct host_alias* a;
 
 	rpc->add(c, "{", &hr);
 	rpc->struct_add(hr, "s",
 			"myself_callbacks", is_check_self_func_list_set()?"yes":"no");
+	rpc->struct_add(hr, "[", "aliases", &hs);
 	for(a=aliases; a; a=a->next) {
-		rpc->struct_add(hr, "{", "alias", &ha);
+		rpc->struct_add(hs, "{", "alias", &ha);
 		rpc->struct_add(ha, "sS",
 				"proto",  proto2a(a->proto),
 				"address", &a->alias
@@ -918,7 +943,7 @@ static void core_sockets_list(rpc_t* rpc, void* c)
 	unsigned short proto;
 
 	proto=PROTO_UDP;
-	rpc->add(c, "{", &hr);
+	rpc->add(c, "[", &hr);
 	do{
 		list=get_sock_info_list(proto);
 		for(si=list?*list:0; si; si=si->next){
@@ -941,7 +966,7 @@ static void core_sockets_list(rpc_t* rpc, void* c)
 				rpc->struct_add(ha, "ss",
 						"proto", get_proto_name(proto),
 						"address", si->name.s);
-				if (!si->flags & SI_IS_IP)
+				if (!(si->flags & SI_IS_IP))
 					rpc->struct_add(ha, "s",
 						"ipaddress", si->address_str.s);
 				rpc->struct_add(ha, "sss",
@@ -953,9 +978,88 @@ static void core_sockets_list(rpc_t* rpc, void* c)
 	} while((proto=next_proto(proto)));
 }
 
+/**
+ *
+ */
+static const char* core_modules_doc[] = {
+	"List loaded modules",    /* Documentation string */
+	0                               /* Method signature(s) */
+};
+
+/**
+ * list listen sockets for SIP server
+ */
+static void core_modules(rpc_t* rpc, void* c)
+{
+	sr_module_t* t;
+
+	for(t = get_loaded_modules(); t; t = t->next) {
+		if (rpc->add(c, "s", t->exports.name) < 0) return;
+	}
+}
+
+/**
+ *
+ */
+static const char* core_ppdefines_doc[] = {
+	"List preprocessor defines",    /* Documentation string */
+	0                               /* Method signature(s) */
+};
+
+/**
+ * list listen sockets for SIP server
+ */
+static void core_ppdefines(rpc_t* rpc, void* c)
+{
+	str *ppdef;
+	int i=0;
+
+	while((ppdef=pp_get_define_name(i))!=NULL) {
+		if (rpc->add(c, "s", ppdef->s) < 0) return;
+		i++;
+	}
+}
+
+/**
+ *
+ */
+static const char* core_ppdefines_full_doc[] = {
+	"List preprocessor defines with full details", /* Documentation string */
+	0                               /* Method signature(s) */
+};
+
+/**
+ * list listen sockets for SIP server
+ */
+static void core_ppdefines_full(rpc_t* rpc, void* c)
+{
+	ksr_ppdefine_t *ppdef;
+	void *vp;
+	int i=0;
+
+	while((ppdef=pp_get_define(i))!=NULL && ppdef->name.s!=NULL) {
+		if (rpc->add(c, "{", &vp) < 0) {
+			rpc->fault(c, 500, "Server Failure");
+			return;
+		}
+		if(ppdef->value.s!=NULL) {
+			rpc->struct_add(vp, "SdS",
+					"name", &ppdef->name,
+					"type", ppdef->dtype,
+					"value", &ppdef->value);
+		} else {
+			rpc->struct_add(vp, "Sds",
+					"name", &ppdef->name,
+					"type", ppdef->dtype,
+					"value", "none");
+		}
+
+		i++;
+	}
+}
 
 /*
- * RPC Methods exported by this module
+ * RPC Methods exported by core
  */
 static rpc_export_t core_rpc_methods[] = {
 	{"system.listMethods",     system_listMethods,     system_listMethods_doc,     RET_ARRAY},
@@ -976,6 +1080,7 @@ static rpc_export_t core_rpc_methods[] = {
 	{"core.uptime",            core_uptime,            core_uptime_doc,            0        },
 	{"core.ps",                core_ps,                core_ps_doc,                RET_ARRAY},
 	{"core.psx",               core_psx,               core_psx_doc,               RET_ARRAY},
+	{"core.psa",               core_psa,               core_psa_doc,               RET_ARRAY},
 	{"core.pwd",               core_pwd,               core_pwd_doc,               RET_ARRAY},
 	{"core.arg",               core_arg,               core_arg_doc,               RET_ARRAY},
 	{"core.kill",              core_kill,              core_kill_doc,              0        },
@@ -985,11 +1090,14 @@ static rpc_export_t core_rpc_methods[] = {
 #endif
 	{"core.tcp_info",          core_tcpinfo,           core_tcpinfo_doc,    0},
 	{"core.tcp_options",       core_tcp_options,       core_tcp_options_doc,0},
-	{"core.tcp_list",          core_tcp_list,          core_tcp_list_doc,0},
+	{"core.tcp_list",          core_tcp_list,          core_tcp_list_doc, RET_ARRAY},
 	{"core.udp4_raw_info",     core_udp4rawinfo,       core_udp4rawinfo_doc,
 		0},
-	{"core.aliases_list",      core_aliases_list,      core_aliases_list_doc,   0},
-	{"core.sockets_list",      core_sockets_list,      core_sockets_list_doc,   0},
+	{"core.aliases_list",      core_aliases_list,      core_aliases_list_doc, 0},
+	{"core.sockets_list",      core_sockets_list,      core_sockets_list_doc, 0},
+	{"core.modules",           core_modules,           core_modules_doc,         RET_ARRAY},
+	{"core.ppdefines",         core_ppdefines,         core_ppdefines_doc,       RET_ARRAY},
+	{"core.ppdefines_full",    core_ppdefines_full,    core_ppdefines_full_doc,  RET_ARRAY},
 #ifdef USE_DNS_CACHE
 	{"dns.mem_info",          dns_cache_mem_info,     dns_cache_mem_info_doc,
 		0	},
@@ -998,7 +1106,7 @@ static rpc_export_t core_rpc_methods[] = {
 	{"dns.debug_all",      dns_cache_debug_all,       dns_cache_debug_all_doc,
 		0	},
 	{"dns.view",               dns_cache_view,        dns_cache_view_doc,
-		0	},
+		RET_ARRAY	},
 	{"dns.lookup",             dns_cache_rpc_lookup,  dns_cache_rpc_lookup_doc,
 		0	},
 	{"dns.delete_all",         dns_cache_delete_all,  dns_cache_delete_all_doc,
@@ -1038,20 +1146,20 @@ static rpc_export_t core_rpc_methods[] = {
 		dns_get_server_state_doc, 0 },
 #endif
 #endif
-#ifdef USE_DST_BLACKLIST
-	{"dst_blacklist.mem_info",  dst_blst_mem_info,     dst_blst_mem_info_doc,
+#ifdef USE_DST_BLOCKLIST
+	{"dst_blocklist.mem_info",  dst_blst_mem_info,     dst_blst_mem_info_doc,
 		0	},
-	{"dst_blacklist.debug",    dst_blst_debug,         dst_blst_debug_doc,
+	{"dst_blocklist.debug",    dst_blst_debug,         dst_blst_debug_doc,
 		0	},
-	{"dst_blacklist.view",     dst_blst_view,         dst_blst_view_doc,
+	{"dst_blocklist.view",     dst_blst_view,         dst_blst_view_doc,
 		0	},
-	{"dst_blacklist.delete_all", dst_blst_delete_all, dst_blst_delete_all_doc,
+	{"dst_blocklist.delete_all", dst_blst_delete_all, dst_blst_delete_all_doc,
 		0	},
-	{"dst_blacklist.add",      dst_blst_add,          dst_blst_add_doc,
+	{"dst_blocklist.add",      dst_blst_add,          dst_blst_add_doc,
 		0	},
-#ifdef USE_DST_BLACKLIST_STATS
-	{"dst_blacklist.stats_get", dst_blst_stats_get, dst_blst_stats_get_doc, 0},
-#endif /* USE_DST_BLACKLIST_STATS */
+#ifdef USE_DST_BLOCKLIST_STATS
+	{"dst_blocklist.stats_get", dst_blst_stats_get, dst_blst_stats_get_doc, 0},
+#endif /* USE_DST_BLOCKLIST_STATS */
 #endif
 	{0, 0, 0, 0}
 };
@@ -1067,8 +1175,7 @@ int register_core_rpcs(void)
 		BUG("failed to register core RPCs\n");
 		goto error;
 	}else if (i>0){
-		ERR("%d duplicate RPCs name detected while registering core RPCs\n",
-			 i);
+		ERR("%d duplicate RPCs name detected while registering core RPCs\n", i);
 		goto error;
 	}
 	return 0;
@@ -1080,21 +1187,14 @@ error:
 
 int rpc_init_time(void)
 {
-	char *t;
-	t=ctime(&up_since);
-	if (strlen(t)+1>=MAX_CTIME_LEN) {
-		ERR("Too long data %d\n", (int)strlen(t));
+	char t[MAX_CTIME_LEN];
+	int len;
+	if (! ctime_r(&up_since, t)) {
+		ERR("Invalid time value\n");
 		return -1;
 	}
 	strcpy(up_since_ctime, t);
-	t = up_since_ctime + strlen(up_since_ctime);
-	while(t>up_since_ctime) {
-		if(*t=='\0' || *t=='\r' || *t=='\n') {
-			*t = '\0';
-		} else {
-			break;
-		}
-		t--;
-	}
+	len = strlen(up_since_ctime);
+	if(len>2 && up_since_ctime[len-1]=='\n') up_since_ctime[len-1]='\0';
 	return 0;
 }
