@@ -114,6 +114,7 @@ extern int version_len;
 
 str _ksr_xavp_via_params = STR_NULL;
 str _ksr_xavp_via_fields = STR_NULL;
+int ksr_local_rport = 0;
 
 /** per process fixup function for global_req_flags.
   * It should be called from the configuration framework.
@@ -596,7 +597,7 @@ static inline int lumps_len(struct sip_msg* msg, struct lump* lumps,
 #define SUBST_LUMP_LEN(subst_l) \
 		switch((subst_l)->u.subst){ \
 			case SUBST_RCV_IP: \
-				if (msg->rcv.bind_address){ \
+				if (msg->rcv.bind_address && STR_WITHVAL(recv_address_str)){ \
 					new_len+=recv_address_str->len; \
 					if (msg->rcv.bind_address->address.af!=AF_INET) \
 						new_len+=2; \
@@ -606,7 +607,7 @@ static inline int lumps_len(struct sip_msg* msg, struct lump* lumps,
 				}; \
 				break; \
 			case SUBST_RCV_PORT: \
-				if (msg->rcv.bind_address){ \
+				if (msg->rcv.bind_address && STR_WITHVAL(recv_port_str)){ \
 					new_len+=recv_port_str->len; \
 				}else{ \
 					/* FIXME */ \
@@ -645,14 +646,14 @@ static inline int lumps_len(struct sip_msg* msg, struct lump* lumps,
 				break; \
 			case SUBST_RCV_ALL: \
 			case SUBST_RCV_ALL_EX: \
-				if (msg->rcv.bind_address){ \
+				if (msg->rcv.bind_address && STR_WITHVAL(recv_address_str)){ \
 					new_len+=recv_address_str->len; \
 					if ((msg->rcv.bind_address->address.af==AF_INET6)\
 							&& (recv_address_str->s[0]!='[')\
 							&& (memchr(recv_address_str->s, ':',\
 								recv_address_str->len)!=NULL))\
 						new_len+=2; \
-					if (recv_port_no!=SIP_PORT){ \
+					if (recv_port_no!=SIP_PORT && STR_WITHVAL(recv_port_str)){ \
 						/* add :port_no */ \
 						new_len+=1+recv_port_str->len; \
 					}\
@@ -989,7 +990,7 @@ void process_lumps( struct sip_msg* msg,
 #define SUBST_LUMP(subst_l) \
 	switch((subst_l)->u.subst){ \
 		case SUBST_RCV_IP: \
-			if (msg->rcv.bind_address){  \
+			if (msg->rcv.bind_address && STR_WITHVAL(recv_address_str)){  \
 				if (msg->rcv.bind_address->address.af!=AF_INET){\
 					new_buf[offset]='['; offset++; \
 				}\
@@ -1005,7 +1006,7 @@ void process_lumps( struct sip_msg* msg,
 			}; \
 			break; \
 		case SUBST_RCV_PORT: \
-			if (msg->rcv.bind_address){  \
+			if (msg->rcv.bind_address && STR_WITHVAL(recv_port_str)){  \
 				memcpy(new_buf+offset, recv_port_str->s, \
 						recv_port_str->len); \
 				offset+=recv_port_str->len; \
@@ -1016,7 +1017,7 @@ void process_lumps( struct sip_msg* msg,
 			break; \
 		case SUBST_RCV_ALL: \
 		case SUBST_RCV_ALL_EX: \
-			if (msg->rcv.bind_address){  \
+			if (msg->rcv.bind_address && STR_WITHVAL(recv_address_str)){  \
 				/* address */ \
 				if ((msg->rcv.bind_address->address.af==AF_INET6)\
 						&& (recv_address_str->s[0]!='[')\
@@ -1034,7 +1035,7 @@ void process_lumps( struct sip_msg* msg,
 					new_buf[offset]=']'; offset++; \
 				}\
 				/* :port */ \
-				if (recv_port_no!=SIP_PORT){ \
+				if (recv_port_no!=SIP_PORT && STR_WITHVAL(recv_port_str)){ \
 					new_buf[offset]=':'; offset++; \
 					memcpy(new_buf+offset, \
 							recv_port_str->s, \
@@ -1970,7 +1971,7 @@ clean:
   * depending on the presence of the BUILD_IN_SHM flag, needs freeing when
   *   done) and sets returned_len or 0 on error.
   */
-char * build_req_buf_from_sip_req( struct sip_msg* msg,
+char * build_req_buf_from_sip_req(struct sip_msg* msg,
 								unsigned int *returned_len,
 								struct dest_info* send_info,
 								unsigned int mode)
@@ -1986,6 +1987,7 @@ char * build_req_buf_from_sip_req( struct sip_msg* msg,
 	unsigned int offset, s_offset, size;
 	struct lump* via_anchor;
 	struct lump* via_lump;
+	struct lump* via_rm;
 	struct lump* via_insert_param;
 	struct lump* path_anchor;
 	struct lump* path_lump;
@@ -2031,11 +2033,22 @@ char * build_req_buf_from_sip_req( struct sip_msg* msg,
 
 	via_anchor=anchor_lump(msg, msg->via1->hdr.s-buf, 0, HDR_VIA_T);
 	if (unlikely(via_anchor==0)) goto error00;
-	line_buf = create_via_hf( &via_len, msg, send_info, &branch);
+	line_buf = create_via_hf(&via_len, msg, send_info, &branch);
 	if (unlikely(!line_buf)){
 		LM_ERR("could not create Via header\n");
 		goto error00;
 	}
+	if(unlikely(mode&BUILD_NEW_LOCAL_VIA)) {
+		/* delete exiting top Via header */
+		via_rm = del_lump(msg, msg->h_via1->name.s - msg->buf,
+				msg->h_via1->len, 0);
+		if (via_rm==0) {
+			LM_ERR("failed to remove exiting Via header\n");
+			goto error00;
+		}
+		/* do not update old Via header anymore */
+		mode |= BUILD_NO_VIA1_UPDATE;
+	}
 after_local_via:
 	if(unlikely(mode&BUILD_NO_VIA1_UPDATE))
 		goto after_update_via1;
@@ -2975,16 +2988,18 @@ char* create_via_hf(unsigned int *len,
 #endif /* USE_TCP || USE_SCTP */
 
 	/* test and add rport parameter to local via - rfc3581 */
-	if(msg && msg->msg_flags&FL_ADD_LOCAL_RPORT) {
+	if((ksr_local_rport) || (msg && (msg->msg_flags&FL_ADD_LOCAL_RPORT))) {
 		/* params so far + ';rport' + '\0' */
 		via = (char*)pkg_malloc(extra_params.len+RPORT_LEN);
 		if(via==0) {
-		        PKG_MEM_ERROR;
+			PKG_MEM_ERROR;
 			if (extra_params.s) pkg_free(extra_params.s);
 			return 0;
 		}
-		if(extra_params.len!=0) {
+		if(extra_params.s!=NULL && extra_params.len>0) {
 			memcpy(via, extra_params.s, extra_params.len);
+		}
+		if(extra_params.s!=NULL) {
 			pkg_free(extra_params.s);
 		}
 		memcpy(via + extra_params.len, RPORT, RPORT_LEN-1);
@@ -3005,8 +3020,10 @@ char* create_via_hf(unsigned int *len,
 				if (extra_params.s) pkg_free(extra_params.s);
 				return 0;
 			}
-			if(extra_params.len != 0) {
+			if(extra_params.s!=NULL && extra_params.len>0) {
 				memcpy(via, extra_params.s, extra_params.len);
+			}
+			if(extra_params.s!=NULL) {
 				pkg_free(extra_params.s);
 			}
 			memcpy(via + extra_params.len, sbuf, slen);
@@ -3025,12 +3042,14 @@ char* create_via_hf(unsigned int *len,
 		if(xparams.len>0) {
 			via = (char*)pkg_malloc(extra_params.len+xparams.len+2);
 			if(via==0) {
-			        PKG_MEM_ERROR;
+				PKG_MEM_ERROR;
 				if (extra_params.s) pkg_free(extra_params.s);
 				return 0;
 			}
-			if(extra_params.len != 0) {
+			if(extra_params.s!=NULL && extra_params.len>0) {
 				memcpy(via, extra_params.s, extra_params.len);
+			}
+			if(extra_params.s!=NULL) {
 				pkg_free(extra_params.s);
 			}
 			/* add ';' between via parameters */
@@ -3251,7 +3270,12 @@ int sip_msg_update_buffer(sip_msg_t *msg, str *obuf)
 		LM_ERR("invalid buffer parameter\n");
 		return -1;
 	}
-
+#ifdef USE_TCP
+	if(tcp_get_clone_rcvbuf()==0) {
+		LM_ERR("tcp clone received buffer not enabled\n");
+		return -1;
+	}
+#endif
 	if(obuf->len >= BUF_SIZE) {
 		LM_ERR("new buffer is too large (%d)\n", obuf->len);
 		return -1;
@@ -3276,6 +3300,7 @@ int sip_msg_update_buffer(sip_msg_t *msg, str *obuf)
 	/* restore msg fields */
 	msg->buf = tmp.buf;
 	msg->id = tmp.id;
+	msg->pid = tmp.pid;
 	msg->rcv = tmp.rcv;
 	msg->set_global_address = tmp.set_global_address;
 	msg->set_global_port = tmp.set_global_port;
