@@ -66,6 +66,8 @@ int via_cnt;
 /* global request flags */
 unsigned int global_req_flags = 0;
 
+int ksr_sip_parser_mode = KSR_SIP_PARSER_MODE_STRICT;
+
 /* returns pointer to next header line, and fill hdr_f ;
  * if at end of header returns pointer to the last crlf  (always buf)*/
 char* get_hdr_field(char* const buf, char* const end, struct hdr_field* const hdr)
@@ -115,7 +117,7 @@ char* get_hdr_field(char* const buf, char* const end, struct hdr_field* const hd
 			via_cnt++;
 			vb=pkg_malloc(sizeof(struct via_body));
 			if (vb==0){
-				ERR("out of memory\n");
+				PKG_MEM_ERROR;
 				goto error;
 			}
 			memset(vb,0,sizeof(struct via_body));
@@ -134,7 +136,7 @@ char* get_hdr_field(char* const buf, char* const end, struct hdr_field* const hd
 		case HDR_CSEQ_T:
 			cseq_b=pkg_malloc(sizeof(struct cseq_body));
 			if (cseq_b==0){
-				ERR("out of memory\n");
+				PKG_MEM_ERROR;
 				goto error;
 			}
 			memset(cseq_b, 0, sizeof(struct cseq_body));
@@ -154,7 +156,7 @@ char* get_hdr_field(char* const buf, char* const end, struct hdr_field* const hd
 		case HDR_TO_T:
 			to_b=pkg_malloc(sizeof(struct to_body));
 			if (to_b==0){
-				ERR("out of memory\n");
+				PKG_MEM_ERROR;
 				goto error;
 			}
 			memset(to_b, 0, sizeof(struct to_body));
@@ -169,7 +171,9 @@ char* get_hdr_field(char* const buf, char* const end, struct hdr_field* const hd
 			hdr->body.len=tmp-hdr->body.s;
 			DBG("<%.*s> [%d]; uri=[%.*s]\n", hdr->name.len, ZSW(hdr->name.s),
 					hdr->body.len, to_b->uri.len, ZSW(to_b->uri.s));
-			DBG("to body [%.*s]\n", to_b->body.len, ZSW(to_b->body.s));
+			DBG("to body (%d)[%.*s], to tag (%d)[%.*s]\n", to_b->body.len,
+					to_b->body.len, ZSW(to_b->body.s), to_b->tag_value.len,
+					to_b->tag_value.len, ZSW(to_b->tag_value.s));
 			break;
 		case HDR_CONTENTLENGTH_T:
 			hdr->body.s=tmp;
@@ -241,6 +245,7 @@ char* get_hdr_field(char* const buf, char* const end, struct hdr_field* const hd
 		case HDR_PAI_T:
 		case HDR_PPI_T:
 		case HDR_REASON_T:
+		case HDR_CALLINFO_T:
 		case HDR_OTHER_T:
 			/* just skip over it */
 			hdr->body.s=tmp;
@@ -250,8 +255,10 @@ char* get_hdr_field(char* const buf, char* const end, struct hdr_field* const hd
 				match=q_memchr(tmp, '\n', end-tmp);
 				if (match){
 					match++;
-				}else {
-					ERR("bad body for <%s>(%d)\n", hdr->name.s, hdr->type);
+				} else {
+					ERR("no eol - bad body for <%.*s> (hdr type: %d) [%.*s]\n",
+							 hdr->name.len, hdr->name.s,
+							hdr->type, ((end-tmp)>128)?128:(int)(end-tmp), tmp);
 					/* abort(); */
 					tmp=end;
 					goto error;
@@ -262,7 +269,8 @@ char* get_hdr_field(char* const buf, char* const end, struct hdr_field* const hd
 			hdr->body.len=match-hdr->body.s;
 			break;
 		default:
-			BUG("unknown header type %d\n", hdr->type);
+			BUG("unknown header type %d [%.*s]\n", hdr->type,
+					 ((end-buf)>128)?128:(int)(end-buf), buf);
 			goto error;
 	}
 	/* jku: if \r covered by current length, shrink it */
@@ -308,18 +316,19 @@ int parse_headers(struct sip_msg* const msg, const hdr_flags_t flags, const int
 	if (unlikely(next)) {
 		orig_flag = msg->parsed_flag;
 		msg->parsed_flag &= ~flags;
-	}else
+	} else {
 		orig_flag=0;
+	}
 
 #ifdef EXTRA_DEBUG
 	DBG("flags=%llx\n", (unsigned long long)flags);
 #endif
-	while( tmp<end && (flags & msg->parsed_flag) != flags){
+	while(tmp<end && (flags & msg->parsed_flag) != flags) {
 		prefetch_loc_r(tmp+64, 1);
 		hf=pkg_malloc(sizeof(struct hdr_field));
 		if (unlikely(hf==0)){
+			PKG_MEM_ERROR;
 			ser_error=E_OUT_OF_MEM;
-			ERR("memory allocation error\n");
 			goto error;
 		}
 		memset(hf,0, sizeof(struct hdr_field));
@@ -327,7 +336,8 @@ int parse_headers(struct sip_msg* const msg, const hdr_flags_t flags, const int
 		rest=get_hdr_field(tmp, end, hf);
 		switch (hf->type){
 			case HDR_ERROR_T:
-				ERR("bad header field [%.*s]\n", (end-tmp>20)?20:(int)(end-tmp), tmp);
+				ERR("bad header field [%.*s]\n",
+						(end-tmp>100)?100:(int)(end-tmp), tmp);
 				goto  error;
 			case HDR_EOH_T:
 				msg->eoh=tmp; /* or rest?*/
@@ -347,7 +357,16 @@ int parse_headers(struct sip_msg* const msg, const hdr_flags_t flags, const int
 				msg->parsed_flag|=HDR_T2F(hf->type);
 				break;
 			case HDR_CALLID_T:
-				if (msg->callid==0) msg->callid=hf;
+				if (msg->callid==0) {
+					msg->callid=hf;
+				} else if(ksr_sip_parser_mode & KSR_SIP_PARSER_MODE_STRICT) {
+					if(IS_SIP(msg)) {
+						LOG(cfg_get(core, core_cfg, sip_parser_log),
+								"duplicate Call-ID header field [%.*s]\n",
+								(end-tmp>100)?100:(int)(end-tmp), tmp);
+						goto  error;
+					}
+				}
 				msg->parsed_flag|=HDR_CALLID_F;
 				break;
 			case HDR_SIPIFMATCH_T:
@@ -355,15 +374,42 @@ int parse_headers(struct sip_msg* const msg, const hdr_flags_t flags, const int
 				msg->parsed_flag|=HDR_SIPIFMATCH_F;
 				break;
 			case HDR_TO_T:
-				if (msg->to==0) msg->to=hf;
+				if (msg->to==0) {
+					msg->to=hf;
+				} else if(ksr_sip_parser_mode & KSR_SIP_PARSER_MODE_STRICT) {
+					if(IS_SIP(msg)) {
+						LOG(cfg_get(core, core_cfg, sip_parser_log),
+								"duplicate To header field [%.*s]\n",
+								(end-tmp>100)?100:(int)(end-tmp), tmp);
+						goto  error;
+					}
+				}
 				msg->parsed_flag|=HDR_TO_F;
 				break;
 			case HDR_CSEQ_T:
-				if (msg->cseq==0) msg->cseq=hf;
+				if (msg->cseq==0) {
+					msg->cseq=hf;
+				} else if(ksr_sip_parser_mode & KSR_SIP_PARSER_MODE_STRICT) {
+					if(IS_SIP(msg)) {
+						LOG(cfg_get(core, core_cfg, sip_parser_log),
+								"duplicate CSeq header field [%.*s]\n",
+								(end-tmp>100)?100:(int)(end-tmp), tmp);
+						goto  error;
+					}
+				}
 				msg->parsed_flag|=HDR_CSEQ_F;
 				break;
 			case HDR_FROM_T:
-				if (msg->from==0) msg->from=hf;
+				if (msg->from==0) {
+					msg->from=hf;
+				} else if(ksr_sip_parser_mode & KSR_SIP_PARSER_MODE_STRICT) {
+					if(IS_SIP(msg)) {
+						LOG(cfg_get(core, core_cfg, sip_parser_log),
+								"duplicate From header field [%.*s]\n",
+								(end-tmp>100)?100:(int)(end-tmp), tmp);
+						goto  error;
+					}
+				}
 				msg->parsed_flag|=HDR_FROM_F;
 				break;
 			case HDR_CONTACT_T:
@@ -371,7 +417,16 @@ int parse_headers(struct sip_msg* const msg, const hdr_flags_t flags, const int
 				msg->parsed_flag|=HDR_CONTACT_F;
 				break;
 			case HDR_MAXFORWARDS_T:
-				if(msg->maxforwards==0) msg->maxforwards=hf;
+				if(msg->maxforwards==0) {
+					msg->maxforwards=hf;
+				} else {
+					if(IS_SIP(msg)) {
+						LOG(cfg_get(core, core_cfg, sip_parser_log),
+								"duplicate Max-Forwards header field [%.*s]\n",
+								(end-tmp>100)?100:(int)(end-tmp), tmp);
+						goto  error;
+					}
+				}
 				msg->parsed_flag|=HDR_MAXFORWARDS_F;
 				break;
 			case HDR_ROUTE_T:
@@ -387,7 +442,16 @@ int parse_headers(struct sip_msg* const msg, const hdr_flags_t flags, const int
 				msg->parsed_flag|=HDR_CONTENTTYPE_F;
 				break;
 			case HDR_CONTENTLENGTH_T:
-				if (msg->content_length==0) msg->content_length = hf;
+				if (msg->content_length==0) {
+					msg->content_length = hf;
+				} else if(ksr_sip_parser_mode & KSR_SIP_PARSER_MODE_STRICT) {
+					if(IS_SIP(msg)) {
+						LOG(cfg_get(core, core_cfg, sip_parser_log),
+								"duplicate Content-Length header field [%.*s]\n",
+								(end-tmp>100)?100:(int)(end-tmp), tmp);
+						goto  error;
+					}
+				}
 				msg->parsed_flag|=HDR_CONTENTLENGTH_F;
 				break;
 			case HDR_AUTHORIZATION_T:
@@ -535,6 +599,9 @@ int parse_headers(struct sip_msg* const msg, const hdr_flags_t flags, const int
 		    case HDR_REASON_T:
 				msg->parsed_flag|=HDR_REASON_F;
 				break;
+			case HDR_CALLINFO_T:
+				msg->parsed_flag|=HDR_CALLINFO_F;
+				break;
 			default:
 				BUG("unknown header type %d\n", hf->type);
 				goto error;
@@ -553,15 +620,25 @@ int parse_headers(struct sip_msg* const msg, const hdr_flags_t flags, const int
 #endif
 		tmp=rest;
 	}
+
 skip:
 	msg->unparsed=tmp;
+	if(msg->headers==NULL) {
+		/* nothing parsed - invalid input sip message */
+		goto error1;
+	}
 	/* restore original flags */
 	msg->parsed_flag |= orig_flag;
 	return 0;
 
 error:
+	if (hf) {
+		clean_hdr_field(hf);
+		pkg_free(hf);
+	}
+
+error1:
 	ser_error=E_BAD_REQ;
-	if (hf) pkg_free(hf);
 	/* restore original flags */
 	msg->parsed_flag |= orig_flag;
 	return -1;
@@ -667,7 +744,7 @@ int parse_msg(char* const buf, const unsigned int len, struct sip_msg* const msg
 
 error:
 	/* more debugging, msg->orig is/should be null terminated*/
-	LOG(cfg_get(core, core_cfg, corelog), "ERROR: parse_msg: message=<%.*s>\n",
+	LOG(cfg_get(core, core_cfg, sip_parser_log), "ERROR: parse_msg: message=<%.*s>\n",
 			(int)msg->len, ZSW(msg->buf));
 	return -1;
 }
@@ -701,10 +778,7 @@ void free_sip_msg(struct sip_msg* const msg)
 	if (msg->body_lumps)  free_lump_list(msg->body_lumps);
 	if (msg->reply_lump)   free_reply_lump(msg->reply_lump);
 	msg_ldata_reset(msg);
-	/* don't free anymore -- now a pointer to a static buffer */
-#	ifdef DYN_BUF
-	pkg_free(msg->buf);
-#	endif
+	/* no free of msg->buf -- a pointer to a static buffer */
 }
 
 /**
@@ -739,9 +813,9 @@ int set_dst_uri(struct sip_msg* const msg, const str* const uri)
 		memcpy(msg->dst_uri.s, uri->s, uri->len);
 		msg->dst_uri.len = uri->len;
 	} else {
-		ptr = (char*)pkg_malloc(uri->len);
+		ptr = (char*)pkg_malloc(uri->len + 1);
 		if (!ptr) {
-			ERR("Not enough memory\n");
+			PKG_MEM_ERROR;
 			return -1;
 		}
 
@@ -749,6 +823,7 @@ int set_dst_uri(struct sip_msg* const msg, const str* const uri)
 		if (msg->dst_uri.s) pkg_free(msg->dst_uri.s);
 		msg->dst_uri.s = ptr;
 		msg->dst_uri.len = uri->len;
+		msg->dst_uri.s[msg->dst_uri.len] = '\0';
 	}
 	return 0;
 }
@@ -780,7 +855,7 @@ int set_path_vector(struct sip_msg* msg, str* path)
 	} else {
 		ptr = (char*)pkg_malloc(path->len);
 		if (!ptr) {
-			ERR("not enough pkg memory\n");
+			PKG_MEM_ERROR;
 			return -1;
 		}
 
@@ -795,11 +870,14 @@ int set_path_vector(struct sip_msg* msg, str* path)
 
 void reset_path_vector(struct sip_msg* const msg)
 {
-	if (msg->path_vec.s) {
-		pkg_free(msg->path_vec.s);
+	if (!shm_address_in(msg->path_vec.s)) {
+		if (msg->path_vec.s)
+			pkg_free(msg->path_vec.s);
+		msg->path_vec.s = 0;
+		msg->path_vec.len = 0;
+	} else {
+		LM_WARN("Found path_vec that is not in pkg mem!\n");
 	}
-	msg->path_vec.s = 0;
-	msg->path_vec.len = 0;
 }
 
 
@@ -820,7 +898,7 @@ int set_instance(struct sip_msg* msg, str* instance)
 	} else {
 		ptr = (char*)pkg_malloc(instance->len);
 		if (!ptr) {
-			ERR("not enough pkg memory for instance\n");
+			PKG_MEM_ERROR;
 			return -1;
 		}
 		memcpy(ptr, instance->s, instance->len);
@@ -859,7 +937,7 @@ int set_ruid(struct sip_msg* msg, str* ruid)
 	} else {
 		ptr = (char*)pkg_malloc(ruid->len);
 		if (!ptr) {
-			ERR("not enough pkg memory for ruid\n");
+			PKG_MEM_ERROR;
 			return -1;
 		}
 		memcpy(ptr, ruid->s, ruid->len);
@@ -898,7 +976,7 @@ int set_ua(struct sip_msg* msg, str* location_ua)
 	} else {
 		ptr = (char*)pkg_malloc(location_ua->len);
 		if (!ptr) {
-			ERR("not enough pkg memory for location_ua\n");
+			PKG_MEM_ERROR;
 			return -1;
 		}
 		memcpy(ptr, location_ua->s, location_ua->len);
@@ -958,7 +1036,7 @@ hdr_field_t* get_hdr_by_name(const sip_msg_t* const msg, const char* const name,
 
 	for(hdr = msg->headers; hdr; hdr = hdr->next) {
 		if(hdr->name.len == name_len && *hdr->name.s==*name
-				&& strncmp(hdr->name.s, name, name_len)==0)
+				&& strncasecmp(hdr->name.s, name, name_len)==0)
 			return hdr;
 	}
 	return NULL;
@@ -971,7 +1049,7 @@ hdr_field_t* next_sibling_hdr_by_name(const hdr_field_t* const hf)
 
 	for(hdr = hf->next; hdr; hdr = hdr->next) {
 		if(hdr->name.len == hf->name.len && *hdr->name.s==*hf->name.s
-				&& strncmp(hdr->name.s, hf->name.s, hf->name.len)==0)
+				&& strncasecmp(hdr->name.s, hf->name.s, hf->name.len)==0)
 			return hdr;
 	}
 	return NULL;
@@ -1098,6 +1176,161 @@ int get_src_uri(sip_msg_t *m, int tmode, str *uri)
 	return 0;
 }
 
+/**
+ * get source proto:ip:port (socket address format)
+ */
+int get_src_address_socket(sip_msg_t *m, str *ssock)
+{
+	static char buf[MAX_URI_SIZE];
+	char* p;
+	str ip, port;
+	int len;
+	str proto;
+
+	if (!ssock || !m) {
+		ERR("invalid parameter value\n");
+		return -1;
+	}
+
+	if(get_valid_proto_string(m->rcv.proto, 1, 0, &proto)<0) {
+		ERR("unknown transport protocol\n");
+		return -1;
+	}
+
+	ip.s = ip_addr2a(&m->rcv.src_ip);
+	ip.len = strlen(ip.s);
+
+	port.s = int2str(m->rcv.src_port, &port.len);
+
+	len = proto.len + 1 + ip.len + 2*(m->rcv.src_ip.af==AF_INET6)+ 1 + port.len;
+
+	if (len+1 >= MAX_URI_SIZE) {
+		ERR("buffer too small\n");
+		return -1;
+	}
+
+	p = buf;
+
+	memcpy(p, proto.s, proto.len);
+	p += proto.len;
+
+	*p++ = ':';
+
+	if (m->rcv.src_ip.af==AF_INET6)
+		*p++ = '[';
+	memcpy(p, ip.s, ip.len);
+	p += ip.len;
+	if (m->rcv.src_ip.af==AF_INET6)
+		*p++ = ']';
+
+	*p++ = ':';
+
+	memcpy(p, port.s, port.len);
+	p += port.len;
+	*p = '\0';
+
+	ssock->s = buf;
+	ssock->len = len;
+
+	return 0;
+}
+
+/**
+ * get received-on-socket ip, port and protocol in SIP URI format
+ * - tmode - 0: short format (transport=udp is not added, being default)
+ * - atype - 0: listen address; 1: advertised address
+ */
+int get_rcv_socket_uri(sip_msg_t *m, int tmode, str *uri, int atype)
+{
+	static char buf[MAX_URI_SIZE];
+	char* p;
+	str ip, port;
+	int len;
+	str proto;
+
+	if (!uri || !m || !m->rcv.bind_address) {
+		ERR("invalid parameter value\n");
+		return -1;
+	}
+
+	if(tmode==0) {
+		switch(m->rcv.proto) {
+			case PROTO_NONE:
+			case PROTO_UDP:
+				proto.s = 0; /* Do not add transport parameter, UDP is default */
+				proto.len = 0;
+			break;
+			default:
+				if(get_valid_proto_string(m->rcv.proto, 1, 0, &proto)<0) {
+					ERR("unknown transport protocol\n");
+					return -1;
+				}
+		}
+	} else {
+		if(get_valid_proto_string(m->rcv.proto, 1, 0, &proto)<0) {
+			ERR("unknown transport protocol\n");
+			return -1;
+		}
+	}
+
+	if(atype==0 || m->rcv.bind_address->useinfo.address_str.len<=0) {
+		ip.s = m->rcv.bind_address->address_str.s;
+		ip.len = m->rcv.bind_address->address_str.len;
+	} else {
+		ip.s = m->rcv.bind_address->useinfo.address_str.s;
+		ip.len = m->rcv.bind_address->useinfo.address_str.len;
+	}
+
+	if(atype==0|| m->rcv.bind_address->useinfo.port_no_str.len <= 0) {
+		port.s = m->rcv.bind_address->port_no_str.s;
+		port.len = m->rcv.bind_address->port_no_str.len;
+	} else {
+		port.s = m->rcv.bind_address->useinfo.port_no_str.s;
+		port.len = m->rcv.bind_address->useinfo.port_no_str.len;
+	}
+
+	len = 4 + ip.len + 2*(m->rcv.src_ip.af==AF_INET6)+ 1 + port.len;
+	if (proto.s) {
+		len += TRANSPORT_PARAM_LEN;
+		len += proto.len;
+	}
+
+	if (len > MAX_URI_SIZE) {
+		ERR("buffer too small\n");
+		return -1;
+	}
+
+	p = buf;
+	memcpy(p, "sip:", 4);
+	p += 4;
+
+	if (m->rcv.src_ip.af==AF_INET6)
+		*p++ = '[';
+	memcpy(p, ip.s, ip.len);
+	p += ip.len;
+	if (m->rcv.src_ip.af==AF_INET6)
+		*p++ = ']';
+
+	*p++ = ':';
+
+	memcpy(p, port.s, port.len);
+	p += port.len;
+
+	if (proto.s) {
+		memcpy(p, TRANSPORT_PARAM, TRANSPORT_PARAM_LEN);
+		p += TRANSPORT_PARAM_LEN;
+
+		memcpy(p, proto.s, proto.len);
+		p += proto.len;
+	}
+
+	uri->s = buf;
+	uri->len = len;
+
+	return 0;
+}
+
+
 /*! \brief returns a pointer to the begining of the msg's body
  */
 char* get_body(sip_msg_t* const msg)
