@@ -42,7 +42,10 @@
 #include "rpc_lookup.h"
 #include "sr_compat.h"
 #include "ppcfg.h"
+#include "fmsg.h"
 #include "async_task.h"
+#include "shm_init.h"
+#include "daemonize.h"
 
 #include <sys/stat.h>
 #include <regex.h>
@@ -109,6 +112,24 @@ unsigned int set_modinit_delay(unsigned int v)
 	return r;
 }
 
+/* shut down phase for instance - kept in shared memory */
+static int *_ksr_shutdown_phase = NULL;
+
+int ksr_shutdown_phase_init(void)
+{
+	if((_ksr_shutdown_phase == NULL) && (shm_initialized())) {
+		_ksr_shutdown_phase = (int*)shm_mallocxz(sizeof(int));
+	}
+	return 0;
+}
+/**
+ * return destroy modules phase state
+ */
+int ksr_shutdown_phase(void)
+{
+	return (_ksr_shutdown_phase)?(*_ksr_shutdown_phase):0;
+}
+
 /* keep state if server is in destroy modules phase */
 static int _sr_destroy_modules_phase = 0;
 
@@ -201,149 +222,71 @@ int register_builtin_modules()
 }
 
 
-
-/** convert cmd exports to current format.
- * @param ver - module interface versions (0 == ser, 1 == kam).
- * @param src - null terminated array of cmd exports
- *              (either ser_cmd_export_t or kam_cmd_export_t, depending
- *               on ver).
- * @param mod - pointer to module exports structure.
- * @return - pkg_malloc'ed null terminated sr_cmd_export_v31_t array with
- *           the converted cmd exports  or 0 on error.
- */
-static sr31_cmd_export_t* sr_cmd_exports_convert(unsigned ver,
-													void* src, void* mod)
-{
-	int i, n;
-	ser_cmd_export_t* ser_cmd;
-	kam_cmd_export_t* kam_cmd;
-	sr31_cmd_export_t* ret;
-
-	ser_cmd = 0;
-	kam_cmd = 0;
-	ret = 0;
-	n = 0;
-	/* count the number of elements */
-	if (ver == 0) {
-		ser_cmd = src;
-		for (; ser_cmd[n].name; n++);
-	} else if (ver == 1) {
-		kam_cmd = src;
-		for (; kam_cmd[n].name; n++);
-	} else goto error; /* unknown interface version */
-	/* alloc & init new array */
-	ret = pkg_malloc(sizeof(*ret)*(n+1));
-	memset(ret, 0, sizeof(*ret)*(n+1));
-	/* convert/copy */
-	for (i=0; i < n; i++) {
-		if (ver == 0) {
-			ret[i].name = ser_cmd[i].name;
-			ret[i].function = ser_cmd[i].function;
-			ret[i].param_no = ser_cmd[i].param_no;
-			ret[i].fixup = ser_cmd[i].fixup;
-			ret[i].free_fixup = 0; /* no present in ser  <= 2.1 */
-			ret[i].flags = ser_cmd[i].flags;
-		} else {
-			ret[i].name = kam_cmd[i].name;
-			ret[i].function = kam_cmd[i].function;
-			ret[i].param_no = kam_cmd[i].param_no;
-			ret[i].fixup = kam_cmd[i].fixup;
-			ret[i].free_fixup = kam_cmd[i].free_fixup;
-			ret[i].flags = kam_cmd[i].flags;
-		}
-		/* 3.1+ specific stuff */
-		ret[i].fixup_flags = 0;
-		ret[i].module_exports = mod;
-		/* fill known free fixups */
-		if (ret[i].fixup && ret[i].free_fixup == 0)
-			ret[i].free_fixup = get_fixup_free(ret[i].fixup);
-	}
-	return ret;
-error:
-	return 0;
-}
-
-
-
 /* registers a module,  register_f= module register  functions
  * returns <0 on error, 0 on success */
-static int register_module(unsigned ver, union module_exports_u* e,
-					char* path, void* handle)
+static int register_module(module_exports_t* e, char* path, void* handle)
 {
 	int ret, i;
 	struct sr_module* mod;
 	char defmod[64];
+	int n = 0;
 
 	ret=-1;
 
 	/* add module to the list */
 	if ((mod=pkg_malloc(sizeof(struct sr_module)))==0){
-		LM_ERR("memory allocation failure\n");
+		PKG_MEM_ERROR;
 		ret=E_OUT_OF_MEM;
 		goto error;
 	}
-	memset(mod,0, sizeof(struct sr_module));
+	memset(mod, 0, sizeof(struct sr_module));
 	mod->path=path;
 	mod->handle=handle;
-	mod->orig_mod_interface_ver=ver;
-	/* convert exports to sr31 format */
-	if (ver == 0) {
-		/* ser <= 3.0 */
-		mod->exports.name = e->v0.name;
-		if (e->v0.cmds) {
-			mod->exports.cmds = sr_cmd_exports_convert(ver, e->v0.cmds, mod);
-			if (mod->exports.cmds == 0) {
-				LM_ERR("failed to convert module command exports to 3.1 format"
-						" for module \"%s\" (%s), interface version %d\n",
-						mod->exports.name, mod->path, ver);
-				ret = E_UNSPEC;
-				goto error;
-			}
-		}
-		mod->exports.params = e->v0.params;
-		mod->exports.init_f = e->v0.init_f;
-		mod->exports.response_f = e->v0.response_f;
-		mod->exports.destroy_f = e->v0.destroy_f;
-		mod->exports.onbreak_f = e->v0.onbreak_f;
-		mod->exports.init_child_f = e->v0.init_child_f;
-		mod->exports.dlflags = 0; /* not used in ser <= 3.0 */
-		mod->exports.rpc_methods = e->v0.rpc_methods;
-		/* the rest are 0, not used in ser */
-	} else if (ver == 1) {
-		/* kamailio <= 3.0 */
-		mod->exports.name = e->v1.name;
-		if (e->v1.cmds) {
-			mod->exports.cmds = sr_cmd_exports_convert(ver, e->v1.cmds, mod);
-			if (mod->exports.cmds == 0) {
-				LM_ERR("failed to convert module command exports to 3.1 format"
-						" for module \"%s\" (%s), interface version %d\n",
-						mod->exports.name, mod->path, ver);
-				ret = E_UNSPEC;
-				goto error;
-			}
-		}
-		mod->exports.params = e->v1.params;
-		mod->exports.init_f = e->v1.init_f;
-		mod->exports.response_f = e->v1.response_f;
-		mod->exports.destroy_f = e->v1.destroy_f;
-		mod->exports.onbreak_f = 0; /* not used in k <= 3.0 */
-		mod->exports.init_child_f = e->v1.init_child_f;
-		mod->exports.dlflags = e->v1.dlflags;
-		mod->exports.rpc_methods = 0; /* not used in k <= 3.0 */
-		mod->exports.stats = e->v1.stats;
-		mod->exports.nn_cmds = e->v1.nn_cmds;
-		mod->exports.items = e->v1.items;
-		mod->exports.procs = e->v1.procs;
-	} else {
-		LM_ERR("unsupported module interface version %d\n", ver);
-		ret = E_UNSPEC;
+
+	/* copy and convert fields */
+	mod->exports.name = e->name;
+
+	mod->exports.dlflags = e->dlflags;
+
+	if(e->cmds) {
+		for (n=0; e->cmds[n].name; n++);
+	}
+	mod->exports.cmds = pkg_malloc(sizeof(ksr_cmd_export_t)*(n+1));
+	if (mod->exports.cmds==0) {
+		PKG_MEM_ERROR;
+		ret=E_OUT_OF_MEM;
 		goto error;
 	}
+	memset(mod->exports.cmds, 0, sizeof(ksr_cmd_export_t)*(n+1));
+	for (i=0; i < n; i++) {
+		mod->exports.cmds[i].name = e->cmds[i].name;
+		mod->exports.cmds[i].function = e->cmds[i].function;
+		mod->exports.cmds[i].param_no = e->cmds[i].param_no;
+		mod->exports.cmds[i].fixup = e->cmds[i].fixup;
+		mod->exports.cmds[i].free_fixup = e->cmds[i].free_fixup;
+		mod->exports.cmds[i].flags = e->cmds[i].flags;
 
-	if (mod->exports.items) {
+		mod->exports.cmds[i].fixup_flags = 0;
+		mod->exports.cmds[i].module_exports = mod;
+		/* fill known free fixups */
+		if (mod->exports.cmds[i].fixup && mod->exports.cmds[i].free_fixup == 0) {
+			mod->exports.cmds[i].free_fixup
+					= get_fixup_free(mod->exports.cmds[i].fixup);
+		}
+	}
+
+	mod->exports.params = e->params;
+	mod->exports.rpc_methods = e->rpc_methods;
+	mod->exports.pv_items = e->pv_items;
+	mod->exports.response_f = e->response_f;
+	mod->exports.init_mod_f = e->init_mod_f;
+	mod->exports.init_child_f = e->init_child_f;
+	mod->exports.destroy_mod_f = e->destroy_mod_f;
+
+	if (mod->exports.pv_items) {
 		/* register module pseudo-variables for kamailio modules */
 		LM_DBG("register PV from: %s\n", mod->exports.name);
-		if (register_pvars_mod(mod->exports.name, mod->exports.items)!=0) {
+		if (register_pvars_mod(mod->exports.name, mod->exports.pv_items)!=0) {
 			LM_ERR("failed to register pseudo-variables for module %s (%s)\n",
 				mod->exports.name, path);
 			ret = E_UNSPEC;
@@ -392,7 +335,10 @@ error:
 	return ret;
 }
 
-static inline int version_control(void *handle, char *path)
+/**
+ * check the version of loaded module
+ */
+int ksr_version_control(void *handle, char *path)
 {
 	char **m_ver;
 	char **m_flags;
@@ -433,45 +379,24 @@ static inline int version_control(void *handle, char *path)
 }
 
 /**
- * \brief load a sr module
+ *  \brief lookup a module and fill the new path
  *
- * tries to load the module specified by mod_path.
- * If mod_path is 'modname' or 'modname.so' then
- *  \<MODS_DIR\>/\<modname\>.so will be tried and if this fails
- *  \<MODS_DIR\>/\<modname\>/\<modname\>.so
- * If mod_path contain a '/' it is assumed to be the
- * path to the module and tried first. If fails and mod_path is not
- * absolute path (not starting with '/') then will try:
- * \<MODS_DIR\>/mod_path
  * @param mod_path path or module name
+ * @param new_path resolved path to the module, must be pkg_free() if != mod_path
  * @return 0 on success , <0 on error
  */
-int load_module(char* mod_path)
+int ksr_locate_module(char *mod_path, char **new_path)
 {
-	void* handle;
-	char* error;
-	mod_register_function mr;
-	union module_exports_u* exp;
-	unsigned* mod_if_ver;
-	struct sr_module* t;
 	struct stat stat_buf;
 	str modname;
 	char* mdir;
 	char* nxt_mdir;
 	char* path;
 	int mdir_len;
-	int len;
-	int dlflags;
-	int new_dlflags;
-	int retries;
 	int path_type;
-	str expref;
-	char exbuf[64];
+	int len;
 
-#ifndef RTLD_NOW
-/* for openbsd */
-#define RTLD_NOW DL_LAZY
-#endif
+	*new_path = NULL;
 	path=mod_path;
 	path_type = 0;
 	modname.s = path;
@@ -494,7 +419,10 @@ int load_module(char* mod_path)
 				/* try path <MODS_DIR>/<modname>.so */
 				path = (char*)pkg_malloc(mdir_len + 1 /* "/" */ +
 									modname.len + 3 /* ".so" */ + 1);
-				if (path==0) goto error;
+				if (path==0) {
+					PKG_MEM_ERROR;
+					goto error;
+				}
 				memcpy(path, mdir, mdir_len);
 				len = mdir_len;
 				if (len != 0 && path[len - 1] != '/'){
@@ -515,7 +443,10 @@ int load_module(char* mod_path)
 						mdir_len + 1 /* "/" */ +
 						modname.len + 1 /* "/" */ +
 						modname.len + 3 /* ".so" */ + 1);
-					if (path==0) goto error;
+					if (path==0) {
+						PKG_MEM_ERROR;
+						goto error;
+					}
 					memcpy(path, mdir, mdir_len);
 					len = mdir_len;
 					if (len != 0 && path[len - 1] != '/') {
@@ -547,7 +478,10 @@ int load_module(char* mod_path)
 					/* try path <MODS_DIR>/mod_path - K compat */
 					path = (char*)pkg_malloc(mdir_len + 1 /* "/" */ +
 									strlen(mod_path) + 1);
-					if (path==0) goto error;
+					if (path==0) {
+						PKG_MEM_ERROR;
+						goto error;
+					}
 					memcpy(path, mdir, mdir_len);
 					len = mdir_len;
 					if (len != 0 && path[len - 1] != '/'){
@@ -572,6 +506,57 @@ int load_module(char* mod_path)
 			goto error;
 		}
 	}
+
+	LM_DBG("found module to load <%s>\n", path);
+	*new_path = path;
+	return 0;
+
+error:
+	if(path!=NULL && path!=mod_path) {
+		pkg_free(path);
+	}
+
+	return -1;
+}
+
+/**
+ * \brief load a sr module
+ *
+ * tries to load the module specified by mod_path.
+ * If mod_path is 'modname' or 'modname.so' then
+ *  \<MODS_DIR\>/\<modname\>.so will be tried and if this fails
+ *  \<MODS_DIR\>/\<modname\>/\<modname\>.so
+ * If mod_path contain a '/' it is assumed to be the
+ * path to the module and tried first. If fails and mod_path is not
+ * absolute path (not starting with '/') then will try:
+ * \<MODS_DIR\>/mod_path
+ * @param mod_path path or module name
+ * @return 0 on success , <0 on error
+ */
+int load_module(char* mod_path)
+{
+	void* handle;
+	char* error;
+	mod_register_function mr;
+	module_exports_t* exp;
+	struct sr_module* t;
+	int dlflags;
+	int new_dlflags;
+	int retries;
+	char* path = NULL;
+	str expref;
+	char exbuf[64];
+	char* mdir;
+
+#ifndef RTLD_NOW
+/* for openbsd */
+#define RTLD_NOW DL_LAZY
+#endif
+
+	if(ksr_locate_module(mod_path, &path)<0) {
+		return -1;
+	}
+
 	LM_DBG("trying to load <%s>\n", path);
 
 	retries=2;
@@ -590,13 +575,8 @@ reload:
 		}
 	}
 	/* version control */
-	if (!version_control(handle, path)) {
-		exit(-1);
-	}
-	mod_if_ver = (unsigned *)dlsym(handle, "module_interface_ver");
-	if ( (error =(char*)dlerror())!=0 ){
-		LM_ERR("no module interface version in module <%s>\n", path );
-		goto error1;
+	if (!ksr_version_control(handle, path)) {
+		ksr_exit(-1);
 	}
 	/* launch register */
 	mr = (mod_register_function)dlsym(handle, "mod_register");
@@ -617,8 +597,11 @@ reload:
 			goto error;
 		}
 	}
-	exp = (union module_exports_u*)dlsym(handle, "exports");
+	exp = (module_exports_t*)dlsym(handle, "exports");
 	if(exp==NULL) {
+		error =(char*)dlerror();
+		LM_DBG("attempt to lookup exports structure failed - dlerror: %s\n",
+				(error)?error:"none");
 		/* 'exports' structure not found, look up for '_modulename_exports' */
 		mdir = strrchr(mod_path, '/');
 		if (!mdir) {
@@ -630,30 +613,29 @@ reload:
 		if(expref.len>3 && strcmp(expref.s+expref.len-3, ".so")==0)
 			expref.len -= 3;
 		snprintf(exbuf, 62, "_%.*s_exports", expref.len, expref.s);
-		exp = (union module_exports_u*)dlsym(handle, exbuf);
 		LM_DBG("looking up exports with name: %s\n", exbuf);
-		if ( (error =(char*)dlerror())!=0 ){
-			LM_ERR("%s\n", error);
+		exp = (module_exports_t*)dlsym(handle, exbuf);
+		if(exp==NULL || (error =(char*)dlerror())!=0 ){
+			LM_ERR("failure for exports symbol: %s - dlerror: %s\n",
+					exbuf, (error)?error:"none");
 			goto error1;
 		}
 	}
 	/* hack to allow for kamailio style dlflags inside exports */
-	if (*mod_if_ver == 1) {
-		new_dlflags = exp->v1.dlflags;
-		if (new_dlflags!=dlflags && new_dlflags!=DEFAULT_DLFLAGS) {
-			/* we have to reload the module */
-			dlclose(handle);
-			DEBUG("%s: exports dlflags interface is deprecated and it will not"
-					" be supported in newer versions; consider using"
-					" mod_register() instead\n", path);
-			dlflags=new_dlflags;
-			retries--;
-			if (retries>0) goto reload;
-			LM_ERR("%s: cannot agree on the dlflags\n", path);
-			goto error;
-		}
+	new_dlflags = exp->dlflags;
+	if (new_dlflags!=dlflags && new_dlflags!=DEFAULT_DLFLAGS) {
+		/* we have to reload the module */
+		dlclose(handle);
+		DEBUG("%s: exports dlflags interface is deprecated and it will not"
+				" be supported in newer versions; consider using"
+				" mod_register() instead\n", path);
+		dlflags=new_dlflags;
+		retries--;
+		if (retries>0) goto reload;
+		LM_ERR("%s: cannot agree on the dlflags\n", path);
+		goto error;
 	}
-	if (register_module(*mod_if_ver, exp, path, handle)<0) goto error1;
+	if (register_module(exp, path, handle)<0) goto error1;
 	return 0;
 
 error1:
@@ -665,20 +647,55 @@ skip:
 	return -1;
 }
 
+/**
+ *
+ */
+int load_modulex(char* mod_path)
+{
+	str seval;
+	str sfmt;
+	sip_msg_t *fmsg;
+	char* emod;
 
+	emod = mod_path;
+	if(strchr(mod_path, '$') != NULL) {
+		fmsg = faked_msg_get_next();
+		sfmt.s = mod_path;
+		sfmt.len = strlen(sfmt.s);
+		if(pv_eval_str(fmsg, &seval, &sfmt)>=0) {
+			emod = seval.s;
+		}
+	}
+
+	return load_module(emod);
+}
+
+/**
+ * test if command flags are compatible with route block flags (type)
+ * - decide if the command is allowed to run within a specific route block
+ * - return: 1 if allowed; 0 if not allowed
+ */
+static inline int sr_cmd_flags_match(int cflags, int rflags)
+{
+	if((cflags & rflags) == rflags) {
+		return 1;
+	}
+	if((rflags==EVENT_ROUTE) && (cflags & EVENT_ROUTE)) {
+		return 1;
+	}
+	return 0;
+}
 
 /* searches the module list for function name in module mod and returns
  *  a pointer to the "name" function record union or 0 if not found
- * sets also *mod_if_ver to the original module interface version.
  * mod==0 is a wildcard matching all modules
  * flags parameter is OR value of all flags that must match
  */
-sr31_cmd_export_t* find_mod_export_record(char* mod, char* name,
-											int param_no, int flags,
-											unsigned* mod_if_ver)
+ksr_cmd_export_t* find_mod_export_record(char* mod, char* name,
+											int param_no, int flags)
 {
 	struct sr_module* t;
-	sr31_cmd_export_t* cmd;
+	ksr_cmd_export_t* cmd;
 
 	for(t=modules;t;t=t->next){
 		if (mod!=0 && (strcmp(t->exports.name, mod) !=0))
@@ -688,16 +705,15 @@ sr31_cmd_export_t* find_mod_export_record(char* mod, char* name,
 				if((strcmp(name, cmd->name) == 0) &&
 					((cmd->param_no == param_no) ||
 					(cmd->param_no==VAR_PARAM_NO)) &&
-					((cmd->flags & flags) == flags)
+					(sr_cmd_flags_match(cmd->flags, flags)==1)
 				){
-					LM_DBG("find_export_record: found <%s> in module %s [%s]\n",
+					LM_DBG("found export of <%s> in module %s [%s]\n",
 						name, t->exports.name, t->path);
-					*mod_if_ver=t->orig_mod_interface_ver;
 					return cmd;
 				}
 			}
 	}
-	LM_DBG("find_export_record: <%s> not found \n", name);
+	LM_DBG("export of <%s> not found (flags %d)\n", name, flags);
 	return 0;
 }
 
@@ -705,26 +721,21 @@ sr31_cmd_export_t* find_mod_export_record(char* mod, char* name,
 
 /* searches the module list for function name and returns
  *  a pointer to the "name" function record union or 0 if not found
- * sets also *mod_if_ver to the module interface version (needed to know
- * which member of the union should be accessed v0 or v1)
  * mod==0 is a wildcard matching all modules
  * flags parameter is OR value of all flags that must match
  */
-sr31_cmd_export_t* find_export_record(char* name,
-											int param_no, int flags,
-											unsigned* mod_if_ver)
+ksr_cmd_export_t* find_export_record(char* name, int param_no, int flags)
 {
-	return find_mod_export_record(0, name, param_no, flags, mod_if_ver);
+	return find_mod_export_record(0, name, param_no, flags);
 }
 
 
 
 cmd_function find_export(char* name, int param_no, int flags)
 {
-	sr31_cmd_export_t* cmd;
-	unsigned mver;
+	ksr_cmd_export_t* cmd;
 
-	cmd = find_export_record(name, param_no, flags, &mver);
+	cmd = find_export_record(name, param_no, flags);
 	return cmd?cmd->function:0;
 }
 
@@ -743,10 +754,9 @@ rpc_export_t* find_rpc_export(char* name, int flags)
  */
 cmd_function find_mod_export(char* mod, char* name, int param_no, int flags)
 {
-	sr31_cmd_export_t* cmd;
-	unsigned mver;
+	ksr_cmd_export_t* cmd;
 
-	cmd=find_mod_export_record(mod, name, param_no, flags, &mver);
+	cmd=find_mod_export_record(mod, name, param_no, flags);
 	if (cmd)
 		return cmd->function;
 
@@ -767,6 +777,9 @@ struct sr_module* find_module_by_name(char* mod) {
 	return 0;
 }
 
+sr_module_t* get_loaded_modules(void) {
+	return modules;
+}
 
 /*!
  * \brief Find a parameter with given type
@@ -803,12 +816,18 @@ void destroy_modules()
 	struct sr_module* t, *foo;
 
 	_sr_destroy_modules_phase = 1;
+	if(_ksr_shutdown_phase!=NULL) {
+		*_ksr_shutdown_phase = 1;
+	}
+
+	LM_DBG("starting modules destroy phase\n");
+
 	/* call first destroy function from each module */
 	t=modules;
 	while(t) {
 		foo=t->next;
-		if (t->exports.destroy_f){
-			t->exports.destroy_f();
+		if (t->exports.destroy_mod_f){
+			t->exports.destroy_mod_f();
 		}
 		t=foo;
 	}
@@ -826,83 +845,6 @@ void destroy_modules()
 	}
 }
 
-#ifdef NO_REVERSE_INIT
-
-/*
- * Initialize all loaded modules, the initialization
- * is done *AFTER* the configuration file is parsed
- */
-int init_modules(void)
-{
-	struct sr_module* t;
-
-	if(async_task_init()<0)
-		return -1;
-
-	for(t = modules; t; t = t->next) {
-		if (t->exports.init_f) {
-			if (t->exports.init_f() != 0) {
-				LM_ERR("Error while initializing module %s\n", t->exports.name);
-				return -1;
-			}
-			/* delay next module init, if configured */
-			if(unlikely(modinit_delay>0))
-				sleep_us(modinit_delay);
-		}
-		if (t->exports.response_f)
-			mod_response_cbk_no++;
-	}
-	mod_response_cbks=pkg_malloc(mod_response_cbk_no *
-									sizeof(response_function));
-	if (mod_response_cbks==0){
-		LM_ERR("memory allocation failure for %d response_f callbacks\n",
-					mod_response_cbk_no);
-		return -1;
-	}
-	for (t=modules, i=0; t && (i<mod_response_cbk_no); t=t->next) {
-		if (t->exports.response_f) {
-			mod_response_cbks[i]=t->exports.response_f;
-			i++;
-		}
-	}
-	return 0;
-}
-
-
-
-/*
- * per-child initialization
- */
-int init_child(int rank)
-{
-	struct sr_module* t;
-	char* type;
-
-	switch(rank) {
-	case PROC_MAIN:     type = "PROC_MAIN";     break;
-	case PROC_TIMER:    type = "PROC_TIMER";    break;
-	case PROC_FIFO:     type = "PROC_FIFO";     break;
-	case PROC_TCP_MAIN: type = "PROC_TCP_MAIN"; break;
-	default:            type = "CHILD";         break;
-	}
-	LM_DBG("initializing %s with rank %d\n", type, rank);
-
-	if(async_task_child_init(rank)<0)
-		return -1;
-
-	for(t = modules; t; t = t->next) {
-		if (t->exports.init_child_f) {
-			if ((t->exports.init_child_f(rank)) < 0) {
-				LM_ERR("Initialization of child %d failed\n", rank);
-				return -1;
-			}
-		}
-	}
-	return 0;
-}
-
-#else
-
 
 /* recursive module child initialization; (recursion is used to
  * process the module linear list in the same order in
@@ -917,10 +859,12 @@ static int init_mod_child( struct sr_module* m, int rank )
 		 */
 		if (init_mod_child(m->next, rank)!=0) return -1;
 		if (m->exports.init_child_f) {
-			LM_DBG("rank %d: %s\n", rank, m->exports.name);
+			LM_DBG("idx %d rank %d: %s [%s]\n", process_no, rank,
+					m->exports.name, my_desc());
 			if (m->exports.init_child_f(rank)<0) {
-				LM_ERR("Error while initializing module %s (%s)\n",
-							m->exports.name, m->path);
+				LM_ERR("error while initializing module %s (%s)"
+						" (idx: %d rank: %d desc: [%s])\n",
+						m->exports.name, m->path, process_no, rank, my_desc());
 				return -1;
 			} else {
 				/* module correctly initialized */
@@ -941,10 +885,32 @@ static int init_mod_child( struct sr_module* m, int rank )
  */
 int init_child(int rank)
 {
+	int ret;
+	char* type;
+
+	switch(rank) {
+	case PROC_MAIN:       type = "PROC_MAIN";       break;
+	case PROC_TIMER:      type = "PROC_TIMER";      break;
+	case PROC_RPC:        type = "PROC_RPC";        break;
+	case PROC_TCP_MAIN:   type = "PROC_TCP_MAIN";   break;
+	case PROC_UNIXSOCK:   type = "PROC_UNIXSOCK";   break;
+	case PROC_ATTENDANT:  type = "PROC_ATTENDANT";  break;
+	case PROC_INIT:       type = "PROC_INIT";       break;
+	case PROC_NOCHLDINIT: type = "PROC_NOCHLDINIT"; break;
+	case PROC_SIPINIT:    type = "PROC_SIPINIT";    break;
+	case PROC_SIPRPC:     type = "PROC_SIPRPC";     break;
+	default:              type = "CHILD";           break;
+	}
+	LM_DBG("initializing %s with rank %d\n", type, rank);
+
 	if(async_task_child_init(rank)<0)
 		return -1;
 
-	return init_mod_child(modules, rank);
+	ret = init_mod_child(modules, rank);
+	if(rank!=PROC_INIT) {
+		pt[process_no].status = 1;
+	}
+	return ret;
 }
 
 
@@ -961,9 +927,9 @@ static int init_mod( struct sr_module* m )
 		 * propagate it up the stack
 		 */
 		if (init_mod(m->next)!=0) return -1;
-			if (m->exports.init_f) {
+			if (m->exports.init_mod_f) {
 				LM_DBG("%s\n", m->exports.name);
-				if (m->exports.init_f()!=0) {
+				if (m->exports.init_mod_f()!=0) {
 					LM_ERR("Error while initializing module %s (%s)\n",
 								m->exports.name, m->path);
 					return -1;
@@ -1002,7 +968,7 @@ int init_modules(void)
 	mod_response_cbks=pkg_malloc(mod_response_cbk_no *
 									sizeof(response_function));
 	if (mod_response_cbks==0){
-		LM_ERR("memory allocation failure for %d response_f callbacks\n", mod_response_cbk_no);
+		PKG_MEM_ERROR;
 		return -1;
 	}
 	for (t=modules, i=0; t && (i<mod_response_cbk_no); t=t->next)
@@ -1014,8 +980,6 @@ int init_modules(void)
 	return 0;
 }
 
-#endif
-
 
 action_u_t *fixup_get_param(void **cur_param, int cur_param_no,
 							int required_param_no)
@@ -1138,7 +1102,7 @@ int fix_param(int type, void** param)
 
 	p = (fparam_t*)pkg_malloc(sizeof(fparam_t));
 	if (!p) {
-		LM_ERR("No memory left\n");
+		PKG_MEM_ERROR;
 		return E_OUT_OF_MEM;
 	}
 	memset(p, 0, sizeof(fparam_t));
@@ -1171,7 +1135,7 @@ int fix_param(int type, void** param)
 			break;
 		case FPARAM_REGEX:
 			if ((p->v.regex = pkg_malloc(sizeof(regex_t))) == 0) {
-				LM_ERR("No memory left\n");
+				PKG_MEM_ERROR;
 				goto error;
 			}
 			if (regcomp(p->v.regex, *param,
@@ -1233,7 +1197,7 @@ int fix_param(int type, void** param)
 			}
 			p->v.pvs=pkg_malloc(sizeof(pv_spec_t));
 			if (p->v.pvs==0){
-				LM_ERR("out of memory while parsing pv_spec_t\n");
+				PKG_MEM_ERROR;
 				goto error;
 			}
 			if (pv_parse_spec2(&name, p->v.pvs, 1)==0){
