@@ -32,7 +32,9 @@
 #include "../../core/strutils.h"
 #include "../../core/tcp_conn.h"
 #include "../../core/pvapi.h"
+#include "../../core/ppcfg.h"
 #include "../../core/trim.h"
+#include "../../core/msg_translator.h"
 
 #include "../../core/parser/parse_from.h"
 #include "../../core/parser/parse_uri.h"
@@ -50,6 +52,8 @@
 #include "pv_core.h"
 #include "pv_svar.h"
 
+#include <string.h>
+#include <stdlib.h>
 
 static str str_udp    = { "UDP", 3 };
 static str str_5060   = { "5060", 4 };
@@ -184,13 +188,30 @@ int pv_get_msgtype(struct sip_msg *msg, pv_param_t *param,
 		return -1;
 
 	if(msg->first_line.type == SIP_REQUEST)
-		type = 1;
+		type = SIP_REQUEST; /* 1 */
 	else if(msg->first_line.type == SIP_REPLY)
-		type = 2;
+		type = SIP_REPLY; /* 2 */
 
 	return pv_get_uintval(msg, param, res, type);
 }
 
+int pv_get_msgtypes(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	char *types = "xx";
+
+	if(msg==NULL)
+		return -1;
+
+	if(msg->first_line.type == SIP_REQUEST) {
+		types = "rq";
+	} else if(msg->first_line.type == SIP_REPLY) {
+		types = "rp";
+	}
+
+	return pv_get_strzval(msg, param, res, types);
+}
+
 int pv_get_version(struct sip_msg *msg, pv_param_t *param,
 		pv_value_t *res)
 {
@@ -302,6 +323,10 @@ int pv_get_xuri_attr(struct sip_msg *msg, struct sip_uri *parsed_uri,
 	} else if(param->pvn.u.isname.name.n==5) /* uri scheme */ {
 		return pv_get_strintval(msg, param, res, &pv_uri_scheme[parsed_uri->type],
 				(int)parsed_uri->type);
+	} else if(param->pvn.u.isname.name.n==6) /* username length */ {
+		if(parsed_uri->user.s==NULL || parsed_uri->user.len<=0)
+			return pv_get_sintval(msg, param, res, 0);
+		return pv_get_sintval(msg, param, res, parsed_uri->user.len);
 	}
 	LM_ERR("unknown specifier\n");
 	return pv_get_null(msg, param, res);
@@ -431,6 +456,11 @@ int pv_get_xto_attr(struct sip_msg *msg, pv_param_t *param,
 			return pv_get_null(msg, param, res);
 		}
 		return pv_get_strval(msg, param, res, &uri->host);
+	} else if(param->pvn.u.isname.name.n==6) /* username length */ {
+		if(uri->user.s==NULL || uri->user.len<=0) {
+			return pv_get_sintval(msg, param, res, 0);
+		}
+		return pv_get_sintval(msg, param, res, uri->user.len);
 	}
 
 	LM_ERR("unknown specifier\n");
@@ -489,6 +519,21 @@ int pv_get_cseq(struct sip_msg *msg, pv_param_t *param,
 	return pv_get_strval(msg, param, res, &(get_cseq(msg)->number));
 }
 
+int pv_get_cseq_body(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	if(msg==NULL)
+		return -1;
+
+	if(msg->cseq==NULL && ((parse_headers(msg, HDR_CSEQ_F, 0)==-1)
+				|| (msg->cseq==NULL)) )
+	{
+		LM_ERR("cannot parse CSEQ header\n");
+		return pv_get_null(msg, param, res);
+	}
+	return pv_get_strval(msg, param, res, &msg->cseq->body);
+}
+
 int pv_get_msg_buf(struct sip_msg *msg, pv_param_t *param,
 		pv_value_t *res)
 {
@@ -501,6 +546,45 @@ int pv_get_msg_buf(struct sip_msg *msg, pv_param_t *param,
 	return pv_get_strval(msg, param, res, &s);
 }
 
+
+static str _ksr_pv_msg_buf_updated = STR_NULL;
+int pv_get_msg_buf_updated(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	dest_info_t send_info;
+
+	if(msg==NULL)
+		return -1;
+
+	if(_ksr_pv_msg_buf_updated.s!=NULL) {
+		pkg_free(_ksr_pv_msg_buf_updated.s);
+		_ksr_pv_msg_buf_updated.s = NULL;
+		_ksr_pv_msg_buf_updated.len = 0;
+	}
+
+	init_dest_info(&send_info);
+	send_info.proto = PROTO_UDP;
+	if(msg->first_line.type == SIP_REPLY) {
+		_ksr_pv_msg_buf_updated.s = generate_res_buf_from_sip_res(msg,
+				(unsigned int*)&_ksr_pv_msg_buf_updated.len,
+				BUILD_NO_VIA1_UPDATE);
+	} else if(msg->first_line.type == SIP_REQUEST) {
+		_ksr_pv_msg_buf_updated.s = build_req_buf_from_sip_req(msg,
+				(unsigned int*)&_ksr_pv_msg_buf_updated.len,
+				&send_info,
+				BUILD_NO_PATH|BUILD_NO_LOCAL_VIA|BUILD_NO_VIA1_UPDATE);
+	} else {
+		return pv_get_null(msg, param, res);
+	}
+
+	if(_ksr_pv_msg_buf_updated.s == NULL) {
+		LM_ERR("couldn't update msg buffer content\n");
+		_ksr_pv_msg_buf_updated.len = 0;
+		return pv_get_null(msg, param, res);
+	}
+	return pv_get_strval(msg, param, res, &_ksr_pv_msg_buf_updated);
+}
+
 int pv_get_msg_len(struct sip_msg *msg, pv_param_t *param,
 		pv_value_t *res)
 {
@@ -531,20 +615,22 @@ int pv_get_flag(struct sip_msg *msg, pv_param_t *param,
 	return pv_get_uintval(msg, param, res, (msg->flags & (1<<param->pvn.u.isname.name.n)) ? 1 : 0);
 }
 
-static inline char* int_to_8hex(int val)
+static inline char* int_to_8hex(int sval)
 {
 	unsigned short digit;
 	int i;
 	static char outbuf[9];
+	unsigned int uval;
 
+	uval = (unsigned int)sval;
 	outbuf[8] = '\0';
 	for(i=0; i<8; i++)
 	{
-		if(val!=0)
+		if(uval!=0)
 		{
-			digit =  val & 0x0f;
+			digit =  uval & 0x0f;
 			outbuf[7-i] = digit >= 10 ? digit + 'a' - 10 : digit + '0';
-			val >>= 4;
+			uval >>= 4;
 		}
 		else
 			outbuf[7-i] = '0';
@@ -663,6 +749,18 @@ int pv_get_srcip(struct sip_msg *msg, pv_param_t *param,
 	return pv_get_strval(msg, param, res, &s);
 }
 
+int pv_get_srcipz(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	str s;
+	if(msg==NULL)
+		return -1;
+
+	s.s = ip_addr2strz(&msg->rcv.src_ip);
+	s.len = strlen(s.s);
+	return pv_get_strval(msg, param, res, &s);
+}
+
 int pv_get_srcport(struct sip_msg *msg, pv_param_t *param,
 		pv_value_t *res)
 {
@@ -697,6 +795,31 @@ int pv_get_srcaddr_uri_helper(struct sip_msg *msg, pv_param_t *param,
 	return pv_get_strval(msg, param, res, &sr);
 }
 
+int pv_get_srcaddr_socket(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	str ssock;
+	str sr;
+
+	if(msg==NULL)
+		return -1;
+
+	if(get_src_address_socket(msg, &ssock)<0)
+		return pv_get_null(msg, param, res);
+
+	if (ssock.len + 1 >= pv_get_buffer_size()) {
+		LM_ERR("local buffer size exceeded\n");
+		return pv_get_null(msg, param, res);
+	}
+
+	sr.s = pv_get_buffer();
+	strncpy(sr.s, ssock.s, ssock.len);
+	sr.len = ssock.len;
+	sr.s[sr.len] = '\0';
+
+	return pv_get_strval(msg, param, res, &sr);
+}
+
 int pv_get_srcaddr_uri(struct sip_msg *msg, pv_param_t *param,
 		pv_value_t *res)
 {
@@ -737,6 +860,144 @@ int pv_get_rcvport(struct sip_msg *msg, pv_param_t *param,
 			&msg->rcv.bind_address->port_no_str);
 }
 
+int pv_get_rcvsname(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	if(msg==NULL)
+		return -1;
+
+	if(msg->rcv.bind_address==NULL
+			|| msg->rcv.bind_address->sockname.s==NULL)
+		return pv_get_null(msg, param, res);
+
+	return pv_get_strval(msg, param, res, &msg->rcv.bind_address->sockname);
+}
+
+int pv_get_rcvaddr_uri_helper(struct sip_msg *msg, pv_param_t *param,
+		int tmode, pv_value_t *res)
+{
+	str uri;
+	str sr;
+
+	if(msg==NULL)
+		return -1;
+
+	if(get_rcv_socket_uri(msg, tmode, &uri, 0)<0)
+		return pv_get_null(msg, param, res);
+
+	if (uri.len + 1 >= pv_get_buffer_size())
+	{
+		LM_ERR("local buffer size exceeded\n");
+		return pv_get_null(msg, param, res);
+	}
+
+	sr.s = pv_get_buffer();
+	strncpy(sr.s, uri.s, uri.len);
+	sr.len = uri.len;
+	sr.s[sr.len] = '\0';
+
+	return pv_get_strval(msg, param, res, &sr);
+}
+
+int pv_get_rcvaddr_uri(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	return pv_get_rcvaddr_uri_helper(msg, param, 0, res);
+}
+
+int pv_get_rcvaddr_uri_full(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	return pv_get_rcvaddr_uri_helper(msg, param, 1, res);
+}
+
+int pv_get_rcv_advertised_ip(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	if(msg==NULL)
+		return -1;
+
+	if(msg->rcv.bind_address!=NULL
+			&& msg->rcv.bind_address->useinfo.address_str.len > 0) {
+		return pv_get_strval(msg, param, res,
+				&msg->rcv.bind_address->useinfo.address_str);
+	}
+
+	return pv_get_rcvip(msg, param, res);
+}
+
+int pv_get_rcv_advertised_port(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	if(msg==NULL)
+		return -1;
+
+	if(msg->rcv.bind_address!=NULL
+			&& msg->rcv.bind_address->useinfo.port_no_str.len > 0) {
+		return pv_get_intstrval(msg, param, res,
+				(int)msg->rcv.bind_address->useinfo.port_no,
+				&msg->rcv.bind_address->useinfo.port_no_str);
+	}
+
+	return pv_get_rcvport(msg, param, res);
+}
+
+int pv_get_rcvadv_uri_helper(struct sip_msg *msg, pv_param_t *param,
+		int tmode, pv_value_t *res)
+{
+	str uri;
+	str sr;
+
+	if(msg==NULL)
+		return -1;
+
+	if(get_rcv_socket_uri(msg, tmode, &uri, 1)<0)
+		return pv_get_null(msg, param, res);
+
+	if (uri.len + 1 >= pv_get_buffer_size())
+	{
+		LM_ERR("local buffer size exceeded\n");
+		return pv_get_null(msg, param, res);
+	}
+
+	sr.s = pv_get_buffer();
+	strncpy(sr.s, uri.s, uri.len);
+	sr.len = uri.len;
+	sr.s[sr.len] = '\0';
+
+	return pv_get_strval(msg, param, res, &sr);
+}
+
+int pv_get_rcvadv_uri(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	if(msg==NULL)
+		return -1;
+
+	if(msg->rcv.bind_address!=NULL
+			&& (msg->rcv.bind_address->useinfo.address_str.len > 0
+				|| msg->rcv.bind_address->useinfo.port_no_str.len > 0)) {
+		return pv_get_rcvadv_uri_helper(msg, param, 0, res);
+	}
+
+	return pv_get_rcvaddr_uri_helper(msg, param, 0, res);
+}
+
+int pv_get_rcvadv_uri_full(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	if(msg==NULL)
+		return -1;
+
+	if(msg->rcv.bind_address!=NULL
+			&& (msg->rcv.bind_address->useinfo.address_str.len > 0
+				|| msg->rcv.bind_address->useinfo.port_no_str.len > 0)) {
+		return pv_get_rcvadv_uri_helper(msg, param, 1, res);
+	}
+
+	return pv_get_rcvaddr_uri_helper(msg, param, 1, res);
+}
+
 /**
  *
  */
@@ -801,6 +1062,34 @@ int pv_get_force_sock(struct sip_msg *msg, pv_param_t *param,
 	return pv_get_strval(msg, param, res, &msg->force_send_socket->sock_str);
 }
 
+int pv_get_force_sock_name(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	if(msg==NULL) {
+		return -1;
+	}
+
+	if (msg->force_send_socket==0
+				|| msg->force_send_socket->sockname.s == NULL) {
+		return pv_get_null(msg, param, res);
+	}
+
+	return pv_get_strval(msg, param, res, &msg->force_send_socket->sockname);
+}
+
+int pv_get_force_sock_port(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	if(msg==NULL)
+		return -1;
+
+	if (msg->force_send_socket==0)
+		return pv_get_uintval(msg, param, res, 0);
+
+	return pv_get_intstrval(msg, param, res, (int)msg->force_send_socket->port_no,
+			&msg->force_send_socket->port_no_str);
+}
+
 int pv_get_useragent(struct sip_msg *msg, pv_param_t *param,
 		pv_value_t *res)
 {
@@ -1098,6 +1387,16 @@ int pv_get_proto(struct sip_msg *msg, pv_param_t *param,
 	return pv_get_strintval(msg, param, res, &s, (int)msg->rcv.proto);
 }
 
+/* proto id of received message */
+int pv_get_protoid(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	if(msg==NULL)
+		return -1;
+
+	return pv_get_sintval(msg, param, res, (int)msg->rcv.proto);
+}
+
 int pv_get_dset(struct sip_msg *msg, pv_param_t *param,
 		pv_value_t *res)
 {
@@ -1105,7 +1404,7 @@ int pv_get_dset(struct sip_msg *msg, pv_param_t *param,
 	if(msg==NULL)
 		return -1;
 
-	s.s = print_dset(msg, &s.len);
+	s.s = print_dset(msg, &s.len, 0);
 	if (s.s == NULL)
 		return pv_get_null(msg, param, res);
 	s.len -= CRLF_LEN;
@@ -1342,6 +1641,37 @@ static inline str *cred_realm(struct sip_msg *rq)
 	return realm;
 }
 
+
+int pv_get_acc_user(struct sip_msg *msg, pv_param_t *param,
+		pv_value_t *res)
+{
+	str* user;
+	struct sip_uri puri;
+	struct to_body* from;
+
+	/* try to take it from credentials */
+	user = cred_user(msg);
+	if (user) {
+		return pv_get_strval(msg, param, res, user);
+	}
+
+	/* from from uri */
+	if(parse_from_header(msg)<0)
+	{
+		LM_ERR("cannot parse FROM header\n");
+		return pv_get_null(msg, param, res);
+	}
+	if (msg->from && (from=get_from(msg)) && from->uri.len) {
+		if (parse_uri(from->uri.s, from->uri.len, &puri) < 0 ) {
+			LM_ERR("bad From URI\n");
+			return pv_get_null(msg, param, res);
+		}
+		return pv_get_strval(msg, param, res, &(puri.user));
+	}
+	return pv_get_null(msg, param, res);
+}
+
+
 int pv_get_acc_username(struct sip_msg *msg, pv_param_t *param,
 		pv_value_t *res)
 {
@@ -1530,7 +1860,7 @@ int pv_get_avp(struct sip_msg *msg,  pv_param_t *param, pv_value_t *res)
 	if ((avp=search_first_avp(name_type, avp_name, &avp_value, &state))==0)
 		return pv_get_null(msg, param, res);
 	res->flags = PV_VAL_STR;
-	if(idxf==0 && idx==0)
+	if(idx==0 && (idxf==PV_IDX_INT || idxf==PV_IDX_NONE))
 	{
 		if(avp->flags & AVP_VAL_STR)
 		{
@@ -1671,6 +2001,8 @@ int pv_get_hdr(struct sip_msg *msg,  pv_param_t *param, pv_value_t *res)
 			if (tv.ri==hf->type)
 				break;
 		} else {
+			if(tv.rs.len==1 && tv.rs.s[0]=='*')
+				break;
 			if (cmp_hdrname_str(&hf->name, &tv.rs)==0)
 				break;
 		}
@@ -1686,7 +2018,7 @@ int pv_get_hdr(struct sip_msg *msg,  pv_param_t *param, pv_value_t *res)
 
 	/* get the value */
 	res->flags = PV_VAL_STR;
-	if(idxf==0 && idx==0)
+	if(idx==0 && (idxf==PV_IDX_INT || idxf==PV_IDX_NONE))
 	{
 		res->rs  = hf->body;
 		return 0;
@@ -1724,8 +2056,10 @@ int pv_get_hdr(struct sip_msg *msg,  pv_param_t *param, pv_value_t *res)
 					if (tv.ri==hf->type)
 						break;
 				} else {
+					if(tv.rs.len==1 && tv.rs.s[0]=='*')
+						break;
 					if (cmp_hdrname_str(&hf->name, &tv.rs)==0)
-					break;
+						break;
 				}
 			}
 		} while (hf);
@@ -1747,8 +2081,11 @@ int pv_get_hdr(struct sip_msg *msg,  pv_param_t *param, pv_value_t *res)
 				if (tv.ri==hf0->type)
 					n++;
 			} else {
-				if (cmp_hdrname_str(&hf0->name, &tv.rs)==0)
+				if(tv.rs.len==1 && tv.rs.s[0]=='*') {
 					n++;
+				} else if (cmp_hdrname_str(&hf0->name, &tv.rs)==0) {
+					n++;
+				}
 			}
 		}
 		idx = -idx;
@@ -1774,8 +2111,11 @@ int pv_get_hdr(struct sip_msg *msg,  pv_param_t *param, pv_value_t *res)
 				if (tv.ri==hf0->type)
 					n++;
 			} else {
-				if (cmp_hdrname_str(&hf0->name, &tv.rs)==0)
+				if(tv.rs.len==1 && tv.rs.s[0]=='*') {
 					n++;
+				} else if (cmp_hdrname_str(&hf0->name, &tv.rs)==0) {
+					n++;
+				}
 			}
 			if(n==idx)
 				break;
@@ -1795,6 +2135,65 @@ int pv_get_hdr(struct sip_msg *msg,  pv_param_t *param, pv_value_t *res)
 
 }
 
+/**
+ *
+ */
+int pv_get_hdrc(struct sip_msg *msg,  pv_param_t *param, pv_value_t *res)
+{
+	pv_value_t tv;
+	struct hdr_field *hf;
+	int hcount;
+
+	if(msg==NULL || res==NULL || param==NULL)
+		return -1;
+
+	hcount = 0;
+
+	/* get the name */
+	if(param->pvn.type == PV_NAME_PVAR)
+	{
+		if(pv_get_spec_name(msg, param, &tv)!=0 || (!(tv.flags&PV_VAL_STR)))
+		{
+			LM_ERR("invalid name\n");
+			return pv_get_sintval(msg, param, res, hcount);
+		}
+	} else {
+		if(param->pvn.u.isname.type == AVP_NAME_STR)
+		{
+			tv.flags = PV_VAL_STR;
+			tv.rs = param->pvn.u.isname.name.s;
+		} else {
+			tv.flags = 0;
+			tv.ri = param->pvn.u.isname.name.n;
+		}
+	}
+	/* we need to be sure we have parsed all headers */
+	if(parse_headers(msg, HDR_EOH_F, 0)<0)
+	{
+		LM_ERR("error parsing headers\n");
+		return pv_get_sintval(msg, param, res, hcount);
+	}
+
+
+	for (hf=msg->headers; hf; hf=hf->next)
+	{
+		if(tv.flags == 0)
+		{
+			if (tv.ri==hf->type) {
+				hcount++;
+			}
+		} else {
+			if (cmp_hdrname_str(&hf->name, &tv.rs)==0) {
+				hcount++;
+			}
+		}
+	}
+	return pv_get_sintval(msg, param, res, hcount);
+}
+
+/**
+ *
+ */
 int pv_get_scriptvar(struct sip_msg *msg,  pv_param_t *param,
 		pv_value_t *res)
 {
@@ -2079,17 +2478,16 @@ int pv_set_ruri(struct sip_msg* msg, pv_param_t *param,
 	memset(&act, 0, sizeof(act));
 	act.val[0].type = STRING_ST;
 	act.val[0].u.string = val->rs.s;
-	backup = val->rs.s[val->rs.len];
-	val->rs.s[val->rs.len] = '\0';
+	STR_VTOZ(val->rs.s[val->rs.len], backup);
 	act.type = SET_URI_T;
 	init_run_actions_ctx(&h);
 	if (do_action(&h, &act, msg)<0)
 	{
 		LM_ERR("do action failed\n");
-		val->rs.s[val->rs.len] = backup;
+		STR_ZTOV(val->rs.s[val->rs.len], backup);
 		goto error;
 	}
-	val->rs.s[val->rs.len] = backup;
+	STR_ZTOV(val->rs.s[val->rs.len], backup);
 
 	return 0;
 error:
@@ -2134,17 +2532,16 @@ int pv_set_ruri_user(struct sip_msg* msg, pv_param_t *param,
 	memset(&act, 0, sizeof(act));
 	act.val[0].type = STRING_ST;
 	act.val[0].u.string = val->rs.s;
-	backup = val->rs.s[val->rs.len];
-	val->rs.s[val->rs.len] = '\0';
+	STR_VTOZ(val->rs.s[val->rs.len], backup);
 	act.type = SET_USER_T;
 	init_run_actions_ctx(&h);
 	if (do_action(&h, &act, msg)<0)
 	{
 		LM_ERR("do action failed\n");
-		val->rs.s[val->rs.len] = backup;
+		STR_ZTOV(val->rs.s[val->rs.len], backup);
 		goto error;
 	}
-	val->rs.s[val->rs.len] = backup;
+	STR_ZTOV(val->rs.s[val->rs.len], backup);
 
 	return 0;
 error:
@@ -2173,17 +2570,16 @@ int pv_set_ruri_host(struct sip_msg* msg, pv_param_t *param,
 	memset(&act, 0, sizeof(act));
 	act.val[0].type = STRING_ST;
 	act.val[0].u.string = val->rs.s;
-	backup = val->rs.s[val->rs.len];
-	val->rs.s[val->rs.len] = '\0';
+	STR_VTOZ(val->rs.s[val->rs.len], backup);
 	act.type = SET_HOST_T;
 	init_run_actions_ctx(&h);
 	if (do_action(&h, &act, msg)<0)
 	{
 		LM_ERR("do action failed\n");
-		val->rs.s[val->rs.len] = backup;
+		STR_ZTOV(val->rs.s[val->rs.len], backup);
 		goto error;
 	}
-	val->rs.s[val->rs.len] = backup;
+	STR_ZTOV(val->rs.s[val->rs.len], backup);
 
 	return 0;
 error:
@@ -2227,17 +2623,16 @@ int pv_set_ruri_port(struct sip_msg* msg, pv_param_t *param,
 	memset(&act, 0, sizeof(act));
 	act.val[0].type = STRING_ST;
 	act.val[0].u.string = val->rs.s;
-	backup = val->rs.s[val->rs.len];
-	val->rs.s[val->rs.len] = '\0';
+	STR_VTOZ(val->rs.s[val->rs.len], backup);
 	act.type = SET_PORT_T;
 	init_run_actions_ctx(&h);
 	if (do_action(&h, &act, msg)<0)
 	{
 		LM_ERR("do action failed\n");
-		val->rs.s[val->rs.len] = backup;
+		STR_ZTOV(val->rs.s[val->rs.len], backup);
 		goto error;
 	}
-	val->rs.s[val->rs.len] = backup;
+	STR_ZTOV(val->rs.s[val->rs.len], backup);
 
 	return 0;
 error:
@@ -2297,15 +2692,15 @@ int pv_set_force_sock(struct sip_msg* msg, pv_param_t *param,
 		goto error;
 	}
 
-	backup = val->rs.s[val->rs.len];
-	val->rs.s[val->rs.len] = '\0';
+	STR_VTOZ(val->rs.s[val->rs.len], backup);
 	if (parse_phostport(val->rs.s, &host.s, &host.len, &port, &proto) < 0)
 	{
 		LM_ERR("invalid socket specification\n");
-		val->rs.s[val->rs.len] = backup;
+		STR_ZTOV(val->rs.s[val->rs.len], backup);
 		goto error;
 	}
-	val->rs.s[val->rs.len] = backup;
+	STR_ZTOV(val->rs.s[val->rs.len], backup);
+	LM_DBG("trying to set send-socket to [%.*s]\n", val->rs.len, val->rs.s);
 	si = grep_sock_info(&host, (unsigned short)port, (unsigned short)proto);
 	if (si!=NULL)
 	{
@@ -2320,6 +2715,40 @@ error:
 	return -1;
 }
 
+int pv_set_force_sock_name(struct sip_msg* msg, pv_param_t *param,
+		int op, pv_value_t *val)
+{
+	struct socket_info *si;
+
+	if(msg==NULL || param==NULL) {
+		LM_ERR("bad parameters\n");
+		return -1;
+	}
+
+	if(val==NULL || (val->flags&PV_VAL_NULL)) {
+		reset_force_socket(msg);
+		return 0;
+	}
+
+	if(!(val->flags&PV_VAL_STR) || val->rs.len<=0) {
+		LM_ERR("str value required to set the force send sock\n");
+		goto error;
+	}
+
+	LM_DBG("trying to set send-socket to name [%.*s]\n", val->rs.len, val->rs.s);
+	si = ksr_get_socket_by_name(&val->rs);
+	if (si!=NULL) {
+		set_force_socket(msg, si);
+	} else {
+		LM_WARN("no socket found to match name [%.*s]\n",
+				val->rs.len, val->rs.s);
+	}
+
+	return 0;
+error:
+	return -1;
+}
+
 int pv_set_mflags(struct sip_msg* msg, pv_param_t *param,
 		int op, pv_value_t *val)
 {
@@ -2826,6 +3255,14 @@ int pv_parse_hdr_name(pv_spec_p sp, str *in)
 		return 0;
 	}
 
+	if(in->len==1 && in->s[0]=='*') {
+		/* match any header name */
+		sp->pvp.pvn.type = PV_NAME_INTSTR;
+		sp->pvp.pvn.u.isname.type = AVP_NAME_STR;
+		sp->pvp.pvn.u.isname.name.s = *in;
+		return 0;
+	}
+
 	if(in->len>=pv_get_buffer_size()-1)
 	{
 		LM_ERR("name too long\n");
@@ -3152,16 +3589,22 @@ int pv_parse_msg_attrs_name(pv_spec_p sp, str *in)
 				sp->pvp.pvn.u.isname.name.n = 2;
 			else if(strncmp(in->s, "hdrs", 4)==0)
 				sp->pvp.pvn.u.isname.name.n = 3;
+			else if(strncmp(in->s, "hdrc", 4)==0)
+				sp->pvp.pvn.u.isname.name.n = 6;
 			else goto error;
 		break;
 		case 5:
 			if(strncmp(in->s, "fline", 5)==0)
 				sp->pvp.pvn.u.isname.name.n = 4;
+			else if(strncmp(in->s, "fpart", 5)==0)
+				sp->pvp.pvn.u.isname.name.n = 7;
 			else goto error;
 		break;
 		case 8:
 			if(strncmp(in->s, "body_len", 8)==0)
 				sp->pvp.pvn.u.isname.name.n = 5;
+			else if(strncmp(in->s, "hdrs_len", 8)==0)
+				sp->pvp.pvn.u.isname.name.n = 8;
 			else goto error;
 		break;
 		default:
@@ -3183,6 +3626,9 @@ error:
 int pv_get_msg_attrs(sip_msg_t *msg, pv_param_t *param, pv_value_t *res)
 {
 	str s;
+	hdr_field_t* hdr;
+	int n;
+
 	if(msg==NULL)
 		return pv_get_null(msg, param, res);
 
@@ -3228,6 +3674,78 @@ int pv_get_msg_attrs(sip_msg_t *msg, pv_param_t *param, pv_value_t *res)
 			if (s.s != NULL)
 				s.len = msg->buf + msg->len - s.s;
 			return pv_get_sintval(msg, param, res, s.len);
+		case 6: /* headers count */
+			n = 0;
+			for(hdr=msg->headers; hdr!=NULL; hdr=hdr->next) {
+				n++;
+			}
+			return pv_get_sintval(msg, param, res, n);
+		case 7: /* first part - first line + headers */
+			if(msg->unparsed==NULL)
+				return pv_get_null(msg, param, res);
+			s.s = msg->buf;
+			s.len = msg->unparsed - s.s;
+			trim(&s);
+			return pv_get_strval(msg, param, res, &s);
+		case 8: /* headers size */
+			if(msg->unparsed==NULL)
+				return pv_get_sintval(msg, param, res, 0);
+			s.s = msg->buf + msg->first_line.len;
+			s.len = msg->unparsed - s.s;
+			trim(&s);
+			return pv_get_sintval(msg, param, res, s.len);
+
+		default:
+			return pv_get_null(msg, param, res);
+	}
+}
+
+/**
+ *
+ */
+int pv_parse_ksr_attrs_name(pv_spec_p sp, str *in)
+{
+	if(sp==NULL || in==NULL || in->len<=0)
+		return -1;
+
+	switch(in->len) {
+		case 3:
+			if(strncmp(in->s, "ver", 3)==0)
+				sp->pvp.pvn.u.isname.name.n = 0;
+			else goto error;
+		break;
+		case 6:
+			if(strncmp(in->s, "verval", 6)==0)
+				sp->pvp.pvn.u.isname.name.n = 1;
+			else goto error;
+		break;
+		default:
+			goto error;
+	}
+	sp->pvp.pvn.type = PV_NAME_INTSTR;
+	sp->pvp.pvn.u.isname.type = 0;
+
+	return 0;
+
+error:
+	LM_ERR("unknown PV ksr key: %.*s\n", in->len, in->s);
+	return -1;
+}
+
+
+/**
+ *
+ */
+int pv_get_ksr_attrs(sip_msg_t *msg, pv_param_t *param, pv_value_t *res)
+{
+	if(param==NULL)
+		return pv_get_null(msg, param, res);
+
+	switch(param->pvn.u.isname.name.n) {
+		case 0: /* version */
+			return pv_get_strzval(msg, param, res, VERSION);
+		case 1: /* version value */
+			return pv_get_uintval(msg, param, res, VERSIONVAL);
 
 		default:
 			return pv_get_null(msg, param, res);
