@@ -29,7 +29,7 @@
 #include "../../core/mem/shm_mem.h"
 #include "../../core/dprint.h"
 #include "../../core/globals.h"
-#include "../../core/md5.h"
+#include "../../core/crypto/md5utils.h"
 #include "../../core/crc.h"
 #include "../../core/ip_addr.h"
 #include "../../core/dset.h"
@@ -60,7 +60,7 @@
 #include "t_fwd.h"
 #endif
 
-#define FROM_TAG_LEN (MD5_LEN + 1 /* - */ + CRC16_LEN) /* length of FROM tags */
+#define FROM_TAG_LEN (MD5_LEN + 1 /* - */ + CRC32_LEN) /* length of FROM tags */
 
 #ifdef WITH_EVENT_LOCAL_REQUEST
 /* where to go for the local request route ("tm:local-request") */
@@ -106,10 +106,16 @@ int uac_init(void)
 /*
  * Generate a From tag
  */
-void generate_fromtag(str* tag, str* callid)
+void generate_fromtag(str* tag, str* callid, str* ruri)
 {
-	     /* calculate from tag from callid */
+	/* calculate from tag from callid and request uri */
 	crcitt_string_array(&from_tag[MD5_LEN + 1], callid, 1);
+	if(ruri) {
+		crcitt_string_array(&from_tag[MD5_LEN + 5], ruri, 1);
+	} else {
+		/* prevent shorter tag in this case, to be changed */
+		crcitt_string_array(&from_tag[MD5_LEN + 5], callid, 1);
+	}
 	tag->s = from_tag;
 	tag->len = FROM_TAG_LEN;
 }
@@ -181,12 +187,12 @@ int uac_refresh_hdr_shortcuts(tm_cell_t *tcell, char *buf, int buf_len)
 	tcell->cseq_n.len = (int)(cs->number.s + cs->number.len - lreq.cseq->name.s);
 
 	LM_DBG("cseq: [%.*s]\n", tcell->cseq_n.len, tcell->cseq_n.s);
-	lreq.buf=0; /* covers the obsolete DYN_BUF */
+	lreq.buf=0;
 	free_sip_msg(&lreq);
 	return 0;
 
 error:
-	lreq.buf=0; /* covers the obsolete DYN_BUF */
+	lreq.buf=0;
 	free_sip_msg(&lreq);
 	return -1;
 }
@@ -221,7 +227,7 @@ static inline int t_run_local_req(
 		uac_req_t *uac_r,
 		struct cell *new_cell, struct retr_buf *request)
 {
-	static struct sip_msg lreq;
+	struct sip_msg lreq = {0};
 	struct onsend_info onsnd_info;
 	tm_xlinks_t backup_xd;
 	int sflag_bk;
@@ -230,9 +236,11 @@ static inline int t_run_local_req(
 	int backup_route_type;
 	struct cell *backup_t;
 	int backup_branch;
-	unsigned int backup_msgid;
+	msg_ctx_id_t backup_ctxid;
 	int refresh_shortcuts = 0;
 	sr_kemi_eng_t *keng = NULL;
+	run_act_ctx_t ra_ctx;
+	run_act_ctx_t *bctx;
 	str evname = str_init("tm:local-request");
 
 	LM_DBG("executing event_route[tm:local-request]\n");
@@ -240,7 +248,7 @@ static inline int t_run_local_req(
 		return -1;
 	}
 	if (unlikely(set_dst_uri(&lreq, uac_r->dialog->hooks.next_hop))) {
-		LM_ERR("failed to set dst_uri");
+		LM_ERR("failed to set dst_uri\n");
 		free_sip_msg(&lreq);
 		return -1;
 	}
@@ -259,27 +267,36 @@ static inline int t_run_local_req(
 	/* set T to the current transaction */
 	backup_t=get_t();
 	backup_branch=get_t_branch();
-	backup_msgid=global_msg_id;
+	backup_ctxid.msgid=tm_global_ctx_id.msgid;
+	backup_ctxid.pid=tm_global_ctx_id.pid;
 	/* fake transaction and message id */
-	global_msg_id=lreq.id;
+	tm_global_ctx_id.msgid=lreq.id;
+	tm_global_ctx_id.pid=lreq.pid;
 	set_t(new_cell, T_BR_UNDEFINED);
+  
+	init_run_actions_ctx(&ra_ctx);
+	
 	if(goto_on_local_req>=0) {
-		run_top_route(event_rt.rlist[goto_on_local_req], &lreq, 0);
+		run_top_route(event_rt.rlist[goto_on_local_req], &lreq, &ra_ctx);
 	} else {
 		keng = sr_kemi_eng_get();
 		if(keng==NULL) {
 			LM_WARN("event callback (%s) set, but no cfg engine\n",
 					tm_event_callback.s);
 		} else {
-			if(keng->froute(&lreq, EVENT_ROUTE,
+			bctx = sr_kemi_act_ctx_get();
+			sr_kemi_act_ctx_set(&ra_ctx);
+			if(sr_kemi_route(keng, &lreq, EVENT_ROUTE,
 						&tm_event_callback, &evname)<0) {
 				LM_ERR("error running event route kemi callback\n");
 			}
+			sr_kemi_act_ctx_set(bctx);
 		}
 	}
 	/* restore original environment */
 	set_t(backup_t, backup_branch);
-	global_msg_id=backup_msgid;
+	tm_global_ctx_id.msgid=backup_ctxid.msgid;
+	tm_global_ctx_id.pid=backup_ctxid.pid;
 	set_route_type( backup_route_type );
 	p_onsend=0;
 
@@ -287,11 +304,21 @@ static inline int t_run_local_req(
 	tm_xdata_swap(new_cell, &backup_xd, 1);
 	setsflagsval(sflag_bk);
 
+	if (unlikely(ra_ctx.run_flags&DROP_R_F)) {
+		LM_DBG("tm:local-request dropped msg. to %.*s\n",
+				lreq.dst_uri.len, lreq.dst_uri.s);
+		refresh_shortcuts = E_DROP;
+		goto clean;
+	}
+
 	/* rebuild the new message content */
-	if(lreq.force_send_socket != uac_r->dialog->send_sock) {
-		LM_DBG("Send socket updated to: %.*s",
-				lreq.force_send_socket->address_str.len,
-				lreq.force_send_socket->address_str.s);
+	if((lreq.force_send_socket != uac_r->dialog->send_sock)
+			|| (lreq.msg_flags&(FL_ADD_LOCAL_RPORT|FL_ADD_SRVID
+					|FL_ADD_XAVP_VIA_PARAMS|FL_USE_XAVP_VIA_FIELDS))) {
+		LM_DBG("local Via update - new socket: [%.*s] - msg-flags: %u",
+			(lreq.force_send_socket)?lreq.force_send_socket->address_str.len:4,
+			(lreq.force_send_socket)?lreq.force_send_socket->address_str.s:"none",
+			lreq.msg_flags);
 
 		/* rebuild local Via - remove previous value
 			* and add the one for the new send socket */
@@ -308,13 +335,16 @@ static inline int t_run_local_req(
 		lreq.add_to_branch_len = lreq.via1->branch->value.len;
 
 		/* update also info about new destination and send sock */
-		uac_r->dialog->send_sock=lreq.force_send_socket;
-		request->dst.send_sock = lreq.force_send_socket;
-		request->dst.proto = lreq.force_send_socket->proto;
+		if(lreq.force_send_socket) {
+			uac_r->dialog->send_sock=lreq.force_send_socket;
+			request->dst.send_sock = lreq.force_send_socket;
+			request->dst.proto = lreq.force_send_socket->proto;
+		}
 
 		LM_DBG("apply new updates with Via to sip msg\n");
 		buf1 = build_req_buf_from_sip_req(&lreq,
-				(unsigned int*)&buf_len1, &request->dst, BUILD_IN_SHM);
+				(unsigned int*)&buf_len1, &request->dst,
+				BUILD_NEW_LOCAL_VIA|BUILD_IN_SHM);
 		if (likely(buf1)){
 			shm_free(*buf);
 			*buf = buf1;
@@ -341,6 +371,7 @@ normal_update:
 		}
 	}
 
+clean:
 	/* clean local msg structure */
 	if (unlikely(lreq.new_uri.s))
 	{
@@ -354,7 +385,7 @@ normal_update:
 		lreq.dst_uri.s=0;
 		lreq.dst_uri.len=0;
 	}
-	lreq.buf=0; /* covers the obsolete DYN_BUF */
+	lreq.buf=0;
 	free_sip_msg(&lreq);
 	return refresh_shortcuts;
 }
@@ -428,6 +459,12 @@ static inline int t_uac_prepare(uac_req_t *uac_r,
 	/* new message => take the dialog send_socket if set, or the default
 	  send_socket if not*/
 	SND_FLAGS_INIT(&snd_flags);
+
+	if (uac_r->dialog->send_sock != NULL)
+	{
+		snd_flags.f |= SND_F_FORCE_SOCKET;
+	}
+
 #ifdef USE_DNS_FAILOVER
 	if ((uri2dst2(cfg_get(core, core_cfg, use_dns_failover) ? &new_cell->uac[0].dns_h : 0,
 			&dst, uac_r->dialog->send_sock, snd_flags,
@@ -448,10 +485,8 @@ static inline int t_uac_prepare(uac_req_t *uac_r,
 		new_cell->flags |= T_IS_INVITE_FLAG;
 		new_cell->flags|=T_AUTO_INV_100 &
 				(!cfg_get(tm, tm_cfg, tm_auto_inv_100) -1);
-#ifdef WITH_AS_SUPPORT
 		if (uac_r->cb_flags & TMCB_DONT_ACK)
 			new_cell->flags |= T_NO_AUTO_ACK;
-#endif
 		lifetime=cfg_get(tm, tm_cfg, tm_max_inv_lifetime);
 	}else
 		lifetime=cfg_get(tm, tm_cfg, tm_max_noninv_lifetime);
@@ -464,11 +499,9 @@ static inline int t_uac_prepare(uac_req_t *uac_r,
 	new_cell->fr_timeout=cfg_get(tm, tm_cfg, fr_timeout);
 	new_cell->fr_inv_timeout=cfg_get(tm, tm_cfg, fr_inv_timeout);
 	new_cell->end_of_life=get_ticks_raw()+lifetime;
-#ifdef TM_DIFF_RT_TIMEOUT
 	/* same as above for retransmission intervals */
 	new_cell->rt_t1_timeout_ms = cfg_get(tm, tm_cfg, rt_t1_timeout_ms);
 	new_cell->rt_t2_timeout_ms = cfg_get(tm, tm_cfg, rt_t2_timeout_ms);
-#endif
 
 	set_kr(REQ_FWDED);
 
@@ -476,10 +509,18 @@ static inline int t_uac_prepare(uac_req_t *uac_r,
 	request->dst = dst;
 	request->flags |= nhtype;
 
-	if (!is_ack) {
-#ifdef TM_DEL_UNREF
-		INIT_REF(new_cell, 1); /* ref'ed only from the hash */
+#ifdef SO_REUSEPORT
+	if (cfg_get(tcp, tcp_cfg, reuse_port)
+			&& request->dst.send_sock->proto == PROTO_TCP) { 
+		if((uac_r->ssockname!=NULL && uac_r->ssockname->len>0)
+				|| (uac_r->ssock!=NULL && uac_r->ssock->len>0)) {
+			request->dst.send_flags.f |= SND_F_FORCE_SOCKET;
+		}
+	}
 #endif
+
+	if (!is_ack) {
+		INIT_REF(new_cell, 1); /* ref'ed only from the hash */
 		hi=dlg2hash(uac_r->dialog);
 		LOCK_HASH(hi);
 		insert_into_hash_table_unsafe(new_cell, hi);
@@ -488,7 +529,7 @@ static inline int t_uac_prepare(uac_req_t *uac_r,
 
 	buf = build_uac_req(uac_r->method, uac_r->headers, uac_r->body, uac_r->dialog, 0, new_cell,
 		&buf_len, &dst);
-	if (!buf) {
+	if (!buf || buf_len<=0) {
 		LM_ERR("Error while building message\n");
 		ret=E_OUT_OF_MEM;
 		goto error1;
@@ -497,6 +538,10 @@ static inline int t_uac_prepare(uac_req_t *uac_r,
 #ifdef WITH_EVENT_LOCAL_REQUEST
 	if (unlikely(goto_on_local_req>=0 || tm_event_callback.len>0)) {
 		refresh_shortcuts = t_run_local_req(&buf, &buf_len, uac_r, new_cell, request);
+		if (unlikely(refresh_shortcuts==E_DROP)) {
+			ret=E_DROP;
+			goto error1;
+		}			
 	}
 #endif
 
@@ -508,7 +553,7 @@ static inline int t_uac_prepare(uac_req_t *uac_r,
 				LM_ERR("failed to parse headers on uas for failover\n");
 			} else {
 				new_cell->uas.request = sip_msg_cloner(&lreq, &sip_msg_len);
-				lreq.buf=0; /* covers the obsolete DYN_BUF */
+				lreq.buf=0;
 				free_sip_msg(&lreq);
 				if (!new_cell->uas.request) {
 					LM_ERR("no more shmem\n");
@@ -568,12 +613,16 @@ static inline int t_uac_prepare(uac_req_t *uac_r,
 	}
 
 error2:
-#ifdef TM_DEL_UNREF
-	if (!is_ack) {
-		UNREF_FREE(new_cell);
-	}else
-#endif
+	if (is_ack) {
 		free_cell(new_cell);
+	} else {
+		if((new_cell->next_c == 0 && new_cell->prev_c == 0)
+				|| (atomic_get_int(&new_cell->ref_count)==0)) {
+			free_cell(new_cell);
+		} else {
+			UNREF_FREE(new_cell, 0);
+		}
+	}
 error3:
 	return ret;
 }
@@ -584,6 +633,7 @@ error3:
 int prepare_req_within(uac_req_t *uac_r,
 		struct retr_buf **dst_req)
 {
+	int ret = -1;
 	if (!uac_r || !uac_r->method || !uac_r->dialog) {
 		LM_ERR("Invalid parameter value\n");
 		goto err;
@@ -598,13 +648,17 @@ int prepare_req_within(uac_req_t *uac_r,
 	if ((uac_r->method->len == 6) && (!memcmp("CANCEL", uac_r->method->s, 6))) goto send;
 	uac_r->dialog->loc_seq.value++; /* Increment CSeq */
  send:
-	return t_uac_prepare(uac_r, dst_req, 0);
+	ret = t_uac_prepare(uac_r, dst_req, 0);
+	
+	if (unlikely(ret < 0 && ret == E_DROP)) {
+		ret = 0;
+	}
 
  err:
 	/* if (cbp) shm_free(cbp); */
 	/* !! never free cbp here because if t_uac_prepare fails, cbp is not freed
 	 * and thus caller has no chance to discover if it is freed or not !! */
-	return -1;
+	return ret;
 }
 
 static inline int send_prepared_request_impl(struct retr_buf *request, int retransmit, int branch)
@@ -690,7 +744,14 @@ int t_uac_with_ids(uac_req_t *uac_r,
 	branch_bm_t added_branches = 1;
 
 	ret = t_uac_prepare(uac_r, &request, &cell);
-	if (ret < 0) return ret;
+	
+	if (ret < 0) {
+		if (unlikely(ret == E_DROP)) {
+			ret = 0;
+		}
+		return ret;
+	}
+
 	is_ack = (uac_r->method->len == 3) && (memcmp("ACK", uac_r->method->s, 3)==0) ? 1 : 0;
 
 	/* equivalent loop to the one in t_forward_nonack */
@@ -707,7 +768,7 @@ int t_uac_with_ids(uac_req_t *uac_r,
 	}
 
 	if (is_ack) {
-		if (cell) free_cell(cell);
+		free_cell(cell);
 		if (ret_index && ret_label)
 			*ret_index = *ret_label = 0;
 	} else {
@@ -719,7 +780,6 @@ int t_uac_with_ids(uac_req_t *uac_r,
 	return ret;
 }
 
-#ifdef WITH_AS_SUPPORT
 struct retr_buf *local_ack_rb(sip_msg_t *rpl_2xx, struct cell *trans,
 					unsigned int branch, str *hdrs, str *body)
 {
@@ -729,22 +789,26 @@ struct retr_buf *local_ack_rb(sip_msg_t *rpl_2xx, struct cell *trans,
 	struct dest_info dst;
 
 	buf_len = (unsigned)sizeof(struct retr_buf);
-	if (! (buffer = build_dlg_ack(rpl_2xx, trans, branch, hdrs, body,
-			&buf_len, &dst))) {
+	buffer = build_dlg_ack(rpl_2xx, trans, branch, hdrs, body,
+			&buf_len, &dst);
+	if (!buffer || buf_len<=0) {
+		if(buffer) {
+			shm_free(buffer);
+		}
 		return 0;
-	} else {
-		/* 'buffer' now points into a contiguous chunk of memory with enough
-		 * room to hold both the retr. buffer and the string raw buffer: it
-		 * points to the begining of the string buffer; we iterate back to get
-		 * the begining of the space for the retr. buffer. */
-		lack = &((struct retr_buf *)buffer)[-1];
-		lack->buffer = buffer;
-		lack->buffer_len = buf_len;
-		lack->dst = dst;
 	}
+	/* 'buffer' now points into a contiguous chunk of memory with enough
+	 * room to hold both the retr. buffer and the string raw buffer: it
+	 * points to the begining of the string buffer; we iterate back to get
+	 * the begining of the space for the retr. buffer. */
+	lack = &((struct retr_buf *)buffer)[-1];
+	lack->buffer = buffer;
+	lack->buffer_len = buf_len;
+	lack->dst = dst;
+
 
 	/* TODO: need next 2? */
-	lack->activ_type = TYPE_LOCAL_ACK;
+	lack->rbtype = TYPE_LOCAL_ACK;
 	lack->my_T = trans;
 
 	return lack;
@@ -811,7 +875,7 @@ int ack_local_uac(struct cell *trans, str *hdrs, str *body)
 
 	if (! (local_ack = local_ack_rb(trans->uac[0].reply, trans, /*branch*/0,
 			hdrs, body))) {
-		LM_ERR("failed to build ACK retransmission buffer");
+		LM_ERR("failed to build ACK retransmission buffer\n");
 		RET_INVALID;
 	} else {
 		/* set the new buffer, but only if not already set (conc. invok.) */
@@ -851,7 +915,6 @@ fin:
 
 #undef RET_INVALID
 }
-#endif /* WITH_AS_SUPPORT */
 
 
 /*
@@ -872,10 +935,14 @@ int req_within(uac_req_t *uac_r)
 		goto err;
 	}
 
-	if(uac_r->ssock!=NULL && uac_r->ssock->len>0
-			&& uac_r->dialog->send_sock==NULL) {
-		/* set local send socket */
-		uac_r->dialog->send_sock = lookup_local_socket(uac_r->ssock);
+	if(uac_r->dialog->send_sock==NULL) {
+		if(uac_r->ssockname!=NULL && uac_r->ssockname->len>0) {
+			/* set local send socket by name */
+			uac_r->dialog->send_sock = ksr_get_socket_by_name(uac_r->ssockname);
+		} else if(uac_r->ssock!=NULL && uac_r->ssock->len>0) {
+			/* set local send socket by address */
+			uac_r->dialog->send_sock = lookup_local_socket(uac_r->ssock);
+		}
 	}
 
 	/* handle alias parameter in uri
@@ -936,7 +1003,7 @@ int req_outside(uac_req_t *uac_r, str* ruri, str* to, str* from, str *next_hop)
 	if (check_params(uac_r, to, from) < 0) goto err;
 
 	generate_callid(&callid);
-	generate_fromtag(&fromtag, &callid);
+	generate_fromtag(&fromtag, &callid, ruri);
 
 	if (new_dlg_uac(&callid, &fromtag, DEFAULT_CSEQ, from, to, &uac_r->dialog) < 0) {
 		LM_ERR("Error while creating new dialog\n");
@@ -952,10 +1019,14 @@ int req_outside(uac_req_t *uac_r, str* ruri, str* to, str* from, str *next_hop)
 	if (next_hop) uac_r->dialog->dst_uri = *next_hop;
 	w_calculate_hooks(uac_r->dialog);
 
-	if(uac_r->ssock!=NULL && uac_r->ssock->len>0
-			&& uac_r->dialog->send_sock==NULL) {
-		/* set local send socket */
-		uac_r->dialog->send_sock = lookup_local_socket(uac_r->ssock);
+	if(uac_r->dialog->send_sock==NULL) {
+		if(uac_r->ssockname!=NULL && uac_r->ssockname->len>0) {
+			/* set local send socket by name */
+			uac_r->dialog->send_sock = ksr_get_socket_by_name(uac_r->ssockname);
+		} else if(uac_r->ssock!=NULL && uac_r->ssock->len>0) {
+			/* set local send socket by address */
+			uac_r->dialog->send_sock = lookup_local_socket(uac_r->ssock);
+		}
 	}
 
 	return t_uac(uac_r);
@@ -983,7 +1054,7 @@ int request(uac_req_t *uac_r, str* ruri, str* to, str* from, str *next_hop)
 	    generate_callid(&callid);
 	else
 	    callid = *uac_r->callid;
-	generate_fromtag(&fromtag, &callid);
+	generate_fromtag(&fromtag, &callid, ruri);
 
 	if (new_dlg_uac(&callid, &fromtag, DEFAULT_CSEQ, from, to, &dialog) < 0) {
 		LM_ERR("Error while creating temporary dialog\n");
@@ -1012,10 +1083,14 @@ int request(uac_req_t *uac_r, str* ruri, str* to, str* from, str *next_hop)
 	 */
 	uac_r->dialog = dialog;
 
-	if(uac_r->ssock!=NULL && uac_r->ssock->len>0
-			&& uac_r->dialog->send_sock==NULL) {
-		/* set local send socket */
-		uac_r->dialog->send_sock = lookup_local_socket(uac_r->ssock);
+	if(uac_r->dialog->send_sock==NULL) {
+		if(uac_r->ssockname!=NULL && uac_r->ssockname->len>0) {
+			/* set local send socket by name */
+			uac_r->dialog->send_sock = ksr_get_socket_by_name(uac_r->ssockname);
+		} else if(uac_r->ssock!=NULL && uac_r->ssock->len>0) {
+			/* set local send socket by address */
+			uac_r->dialog->send_sock = lookup_local_socket(uac_r->ssock);
+		}
 	}
 
 	res = t_uac(uac_r);
