@@ -37,13 +37,15 @@
 #include "../../core/hashes.h"
 #include "../../core/locking.h"
 #include "../../core/trim.h"
+#include "../../core/xavp.h"
 #include "../../core/parser/parse_uri.h"
 #include "../../core/parser/contact/parse_contact.h"
 #include "../../core/parser/parse_from.h"
 #include "../../core/parser/parse_to.h"
+#include "../../core/parser/parse_expires.h"
 
 #include "../../lib/srdb1/db.h"
-#include "../../lib/srutils/sruid.h"
+#include "../../core/utils/sruid.h"
 
 #include "tps_storage.h"
 #include "api.h"
@@ -53,6 +55,17 @@ extern sruid_t _tps_sruid;
 extern db1_con_t* _tps_db_handle;
 extern db_func_t _tpsdbf;
 
+extern str _tps_contact_host;
+extern int _tps_contact_mode;
+extern str _tps_cparam_name;
+extern str _tps_xavu_cfg;
+extern str _tps_xavu_field_acontact;
+extern str _tps_xavu_field_bcontact;
+extern str _tps_xavu_field_contact_host;
+
+extern str _tps_context_param;
+extern str _tps_context_value;
+
 #define TPS_STORAGE_LOCK_SIZE	1<<9
 static gen_lock_set_t *_tps_storage_lock_set = NULL;
 
@@ -63,8 +76,11 @@ int tps_db_insert_dialog(tps_data_t *td);
 int tps_db_clean_dialogs(void);
 int tps_db_insert_branch(tps_data_t *td);
 int tps_db_clean_branches(void);
-int tps_db_load_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd);
+int tps_db_load_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd,
+		uint32_t mode);
 int tps_db_load_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd);
+int tps_db_update_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd,
+		uint32_t mode);
 int tps_db_update_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd,
 		uint32_t mode);
 int tps_db_end_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd);
@@ -79,6 +95,7 @@ static tps_storage_api_t _tps_storage_api = {
 	.clean_branches = tps_db_clean_branches,
 	.load_branch = tps_db_load_branch,
 	.load_dialog = tps_db_load_dialog,
+	.update_branch = tps_db_update_branch,
 	.update_dialog = tps_db_update_dialog,
 	.end_dialog = tps_db_end_dialog
 };
@@ -199,11 +216,15 @@ int tps_storage_branch_rm(sip_msg_t *msg, tps_data_t *td)
 /**
  *
  */
-int tps_storage_fill_contact(sip_msg_t *msg, tps_data_t *td, str *uuid, int dir)
+int tps_storage_fill_contact(sip_msg_t *msg, tps_data_t *td, str *uuid, int dir,
+		int ctmode)
 {
 	str sv;
-	sip_uri_t puri;
+	sip_uri_t puri, curi;
 	int i;
+	int contact_len;
+	int cparam_len;
+	sr_xavp_t *vavu = NULL;
 
 	if(dir==TPS_DIR_DOWNSTREAM) {
 		sv = td->bs_contact;
@@ -215,14 +236,27 @@ int tps_storage_fill_contact(sip_msg_t *msg, tps_data_t *td, str *uuid, int dir)
 		return 0;
 	}
 
-	if(td->cp + 8 + (2*uuid->len) + sv.len >= td->cbuf + TPS_DATA_SIZE) {
-		LM_ERR("insufficient data buffer\n");
-		return -1;
-	}
 	if (parse_uri(sv.s, sv.len, &puri) < 0) {
 		LM_ERR("failed to parse the uri\n");
 		return -1;
 	}
+
+	contact_len = sv.len;
+	if (_tps_contact_host.len)
+		contact_len = sv.len - puri.host.len + _tps_contact_host.len;
+
+	if (ctmode == 1 || ctmode == 2) {
+		cparam_len = _tps_cparam_name.len;
+	} else {
+		cparam_len = 0;
+	}
+
+	if(td->cp + 8 + (2*uuid->len) + cparam_len + contact_len >= td->cbuf
+			+ TPS_DATA_SIZE) {
+		LM_ERR("insufficient data buffer\n");
+		return -1;
+	}
+	/* copy uuid */
 	if(dir==TPS_DIR_DOWNSTREAM) {
 		td->b_uuid.s = td->cp;
 		*td->cp = 'b';
@@ -242,44 +276,171 @@ int tps_storage_fill_contact(sip_msg_t *msg, tps_data_t *td, str *uuid, int dir)
 
 		td->as_contact.s = td->cp;
 	}
+
 	*td->cp = '<';
 	td->cp++;
+	/* look for sip: */
 	for(i=0; i<sv.len; i++) {
 		*td->cp = sv.s[i];
 		td->cp++;
 		if(sv.s[i]==':') break;
 	}
-	if(dir==TPS_DIR_DOWNSTREAM) {
-		*td->cp = 'b';
-	} else {
-		*td->cp = 'a';
-	}
-	td->cp++;
-	memcpy(td->cp, uuid->s, uuid->len);
-	td->cp += uuid->len;
-	*td->cp = '@';
-	td->cp++;
-	memcpy(td->cp, puri.host.s, puri.host.len);
-	td->cp += puri.host.len;
-	if(puri.port.len>0) {
-		*td->cp = ':';
+	if (ctmode == 1 || ctmode == 2) {
+		/* create new URI parameter for Contact header */
+		if (ctmode == 1) {
+			if (dir==TPS_DIR_DOWNSTREAM) {
+				/* extract the contact address */
+				if(parse_headers(msg, HDR_CONTACT_F, 0) < 0 || msg->contact==NULL) {
+					LM_WARN("bad sip message or missing Contact hdr\n");
+					return -1;
+				} else {
+					if(parse_contact(msg->contact)<0
+							|| ((contact_body_t*)msg->contact->parsed)->contacts==NULL
+							|| ((contact_body_t*)msg->contact->parsed)->contacts->next!=NULL) {
+						LM_ERR("bad Contact header\n");
+						return -1;
+					} else {
+						if (parse_uri(((contact_body_t*)msg->contact->parsed)->contacts->uri.s,
+							((contact_body_t*)msg->contact->parsed)->contacts->uri.len, &curi) < 0) {
+							LM_ERR("failed to parse the contact uri\n");
+							return -1;
+						}
+					}
+				}
+				memcpy(td->cp, curi.user.s, curi.user.len);
+				td->cp += curi.user.len;
+			} else {
+				/* extract the ruri */
+				if(parse_sip_msg_uri(msg)<0) {
+					LM_ERR("failed to parse r-uri\n");
+					return -1;
+				}
+				if(msg->parsed_uri.user.len==0) {
+					LM_ERR("no r-uri user\n");
+					return -1;
+				}
+				memcpy(td->cp, msg->parsed_uri.user.s, msg->parsed_uri.user.len);
+				td->cp += msg->parsed_uri.user.len;
+			}
+		} else if (ctmode == 2) {
+			if (dir==TPS_DIR_DOWNSTREAM) {
+				/* extract the a contact */
+				vavu = xavu_get_child_with_sval(&_tps_xavu_cfg,
+							&_tps_xavu_field_acontact);
+				if(vavu==NULL || vavu->val.v.s.len<=0) {
+					LM_ERR("could not evaluate a_contact xavu\n");
+					return -1;
+				}
+				memcpy(td->cp, vavu->val.v.s.s, vavu->val.v.s.len);
+				td->cp += vavu->val.v.s.len;
+			} else {
+				/* extract the b contact */
+				vavu = xavu_get_child_with_sval(&_tps_xavu_cfg,
+							&_tps_xavu_field_bcontact);
+				if(vavu==NULL || vavu->val.v.s.len<=0) {
+					LM_ERR("could not evaluate b_contact xavu\n");
+					return -1;
+				}
+				memcpy(td->cp, vavu->val.v.s.s, vavu->val.v.s.len);
+				td->cp += vavu->val.v.s.len;
+			}
+		}
+
+		if (!((ctmode == 1) && (dir==TPS_DIR_DOWNSTREAM)
+					&& (curi.user.len <= 0))) {
+			*td->cp = '@';
+			td->cp++;
+		}
+
+		/* contact_host xavu takes preference */
+		if (_tps_xavu_cfg.len>0 && _tps_xavu_field_contact_host.len>0) {
+			vavu = xavu_get_child_with_sval(&_tps_xavu_cfg,
+					&_tps_xavu_field_contact_host);
+		}
+		if(vavu!=NULL && vavu->val.v.s.len>0) {
+			memcpy(td->cp, vavu->val.v.s.s, vavu->val.v.s.len);
+			td->cp += vavu->val.v.s.len;
+		} else {
+			if (_tps_contact_host.len) {
+				/* using configured hostname in the contact header */
+				memcpy(td->cp, _tps_contact_host.s, _tps_contact_host.len);
+				td->cp += _tps_contact_host.len;
+			} else {
+				memcpy(td->cp, puri.host.s, puri.host.len);
+				td->cp += puri.host.len;
+			}
+		}
+		if(puri.port.len>0) {
+			*td->cp = ':';
+			td->cp++;
+			memcpy(td->cp, puri.port.s, puri.port.len);
+			td->cp += puri.port.len;
+		}
+		if(puri.transport_val.len>0) {
+			memcpy(td->cp, ";transport=", 11);
+			td->cp += 11;
+			memcpy(td->cp, puri.transport_val.s, puri.transport_val.len);
+			td->cp += puri.transport_val.len;
+		}
+
+		*td->cp = ';';
 		td->cp++;
-		memcpy(td->cp, puri.port.s, puri.port.len);
-		td->cp += puri.port.len;
-	}
-	if(puri.transport_val.len>0) {
-		memcpy(td->cp, ";transport=", 11);
-		td->cp += 11;
-		memcpy(td->cp, puri.transport_val.s, puri.transport_val.len);
-		td->cp += puri.transport_val.len;
+		memcpy(td->cp, _tps_cparam_name.s, _tps_cparam_name.len);
+		td->cp += _tps_cparam_name.len;
+		*td->cp = '=';
+		td->cp++;
+		if(dir==TPS_DIR_DOWNSTREAM) {
+			*td->cp = 'b';
+		} else {
+			*td->cp = 'a';
+		}
+		td->cp++;
+		memcpy(td->cp, uuid->s, uuid->len);
+		td->cp += uuid->len;
+
+	} else {
+		/* create new user part for Contact header URI */
+		if(dir==TPS_DIR_DOWNSTREAM) {
+			*td->cp = 'b';
+		} else {
+			*td->cp = 'a';
+		}
+		td->cp++;
+		memcpy(td->cp, uuid->s, uuid->len);
+		td->cp += uuid->len;
+		*td->cp = '@';
+		td->cp++;
+
+		if (_tps_contact_host.len) {
+			/* using configured hostname in the contact header */
+			memcpy(td->cp, _tps_contact_host.s, _tps_contact_host.len);
+			td->cp += _tps_contact_host.len;
+		} else {
+			memcpy(td->cp, puri.host.s, puri.host.len);
+			td->cp += puri.host.len;
+		}
+		if(puri.port.len>0) {
+			*td->cp = ':';
+			td->cp++;
+			memcpy(td->cp, puri.port.s, puri.port.len);
+			td->cp += puri.port.len;
+		}
+		if(puri.transport_val.len>0) {
+			memcpy(td->cp, ";transport=", 11);
+			td->cp += 11;
+			memcpy(td->cp, puri.transport_val.s, puri.transport_val.len);
+			td->cp += puri.transport_val.len;
+		}
 	}
 
 	*td->cp = '>';
 	td->cp++;
 	if(dir==TPS_DIR_DOWNSTREAM) {
 		td->bs_contact.len = td->cp - td->bs_contact.s;
+		LM_DBG("td->bs %.*s\n",  td->bs_contact.len,  td->bs_contact.s);
 	} else {
 		td->as_contact.len = td->cp - td->as_contact.s;
+		LM_DBG("td->as %.*s\n",  td->as_contact.len,  td->as_contact.s);
 	}
 	return 0;
 }
@@ -329,10 +490,22 @@ int tps_storage_link_msg(sip_msg_t *msg, tps_data_t *td, int dir)
 
 	/* extract the contact address */
 	if(parse_headers(msg, HDR_CONTACT_F, 0)<0 || msg->contact==NULL) {
-		if(td->s_method_id != METHOD_INVITE) {
-			/* no mandatory contact unless is INVITE - done */
+		if((td->s_method_id != METHOD_INVITE)
+				&& (td->s_method_id != METHOD_SUBSCRIBE)){
+			/* no mandatory contact unless is INVITE or SUBSCRIBE - done */
 			return 0;
 		}
+		if(msg->first_line.type==SIP_REPLY) {
+			if(msg->first_line.u.reply.statuscode>=100
+					&& msg->first_line.u.reply.statuscode<200) {
+				/* provisional response with no mandatory contact header */
+				return 0;
+			}
+			if(msg->first_line.u.reply.statuscode>=400) {
+				/* failure response with no mandatory contact header */
+				return 0;
+			}
+		}
 		LM_ERR("bad sip message or missing Contact hdr\n");
 		goto error;
 	}
@@ -342,6 +515,7 @@ int tps_storage_link_msg(sip_msg_t *msg, tps_data_t *td, int dir)
 		LM_ERR("bad Contact header\n");
 		return -1;
 	}
+
 	if(msg->first_line.type==SIP_REQUEST) {
 		if(dir==TPS_DIR_DOWNSTREAM) {
 			td->a_contact = ((contact_body_t*)msg->contact->parsed)->contacts->uri;
@@ -356,6 +530,19 @@ int tps_storage_link_msg(sip_msg_t *msg, tps_data_t *td, int dir)
 		}
 	}
 
+	if  (td->s_method_id == METHOD_SUBSCRIBE) {
+		if(msg->expires && (msg->expires->body.len > 0) && (msg->expires->parsed
+					|| (parse_expires(msg->expires) >= 0))) {
+			td->expires = ((exp_body_t *)msg->expires->parsed)->val;
+		}
+	}
+
+
+	LM_DBG("downstream: %s - acontact: [%.*s] - bcontact: [%.*s]\n",
+			(dir==TPS_DIR_DOWNSTREAM)?"yes":"no",
+			td->a_contact.len, (td->a_contact.len>0)?td->a_contact.s:"",
+			td->b_contact.len, (td->b_contact.len>0)?td->b_contact.s:"");
+
 	return 0;
 
 error:
@@ -365,24 +552,51 @@ error:
 /**
  *
  */
-int tps_storage_record(sip_msg_t *msg, tps_data_t *td, int dialog)
+int tps_storage_record(sip_msg_t *msg, tps_data_t *td, int dialog, int dir)
 {
-	int ret;
+	int ret = -1; /* error if dialog == 0 */
+	str suid;
+	str *sx = NULL;
 
-	sruid_next(&_tps_sruid);
+	if(_tps_context_value.len>0) {
+		sx = &_tps_context_value;
+	} else if(_tps_context_param.len>0) {
+		sx = &_tps_context_param;
+	}
 
-	ret = tps_storage_fill_contact(msg, td, &_tps_sruid.uid, TPS_DIR_DOWNSTREAM);
+	if(dialog==0) {
+		sruid_nextx(&_tps_sruid, sx);
+		suid = _tps_sruid.uid;
+	} else {
+		if(td->a_uuid.len>0) {
+			suid = td->a_uuid;
+		} else if(td->b_uuid.len>0) {
+			suid = td->b_uuid;
+		} else {
+			goto error;
+		}
+		suid.s++;
+		suid.len--;
+	}
+
+	ret = tps_storage_fill_contact(msg, td, &suid, TPS_DIR_DOWNSTREAM,
+			_tps_contact_mode);
 	if(ret<0) goto error;
-	ret = tps_storage_fill_contact(msg, td, &_tps_sruid.uid, TPS_DIR_UPSTREAM);
+	ret = tps_storage_fill_contact(msg, td, &suid, TPS_DIR_UPSTREAM,
+			_tps_contact_mode);
 	if(ret<0) goto error;
-	ret = tps_storage_link_msg(msg, td, TPS_DIR_DOWNSTREAM);
+
+	ret = tps_storage_link_msg(msg, td, dir);
+	if(ret<0) goto error;
+	ret = _tps_storage_api.insert_branch(td);
 	if(ret<0) goto error;
 	if(dialog==0) {
+		if(td->as_contact.len <= 0 && td->bs_contact.len <= 0) {
+			LM_WARN("no local address - do record routing for all initial requests\n");
+		}
 		ret = _tps_storage_api.insert_dialog(td);
 		if(ret<0) goto error;
 	}
-	ret = _tps_storage_api.insert_branch(td);
-	if(ret<0) goto error;
 
 	return 0;
 
@@ -417,6 +631,7 @@ str td_col_a_srcaddr = str_init("a_srcaddr");
 str td_col_b_srcaddr = str_init("b_srcaddr");
 str td_col_s_method = str_init("s_method");
 str td_col_s_cseq = str_init("s_cseq");
+str td_col_x_context = str_init("x_context");
 
 str tt_table_name = str_init("topos_t");
 str tt_col_rectime = str_init("rectime");
@@ -430,11 +645,18 @@ str tt_col_x_rr = str_init("x_rr");
 str tt_col_y_rr = str_init("y_rr");
 str tt_col_s_rr = str_init("s_rr");
 str tt_col_x_uri = str_init("x_uri");
+str tt_col_a_contact = str_init("a_contact");
+str tt_col_b_contact = str_init("b_contact");
+str tt_col_as_contact = str_init("as_contact");
+str tt_col_bs_contact = str_init("bs_contact");
 str tt_col_x_tag = str_init("x_tag");
+str tt_col_a_tag = str_init("a_tag");
+str tt_col_b_tag = str_init("b_tag");
 str tt_col_s_method = str_init("s_method");
 str tt_col_s_cseq = str_init("s_cseq");
+str tt_col_x_context = str_init("x_context");
 
-#define TPS_NR_KEYS	32
+#define TPS_NR_KEYS	48
 
 str _tps_empty = str_init("");
 
@@ -448,6 +670,11 @@ int tps_db_insert_dialog(tps_data_t *td)
 	db_val_t db_vals[TPS_NR_KEYS];
 	int nr_keys;
 
+	if (_tps_db_handle == NULL) {
+		LM_ERR("No database handle - misconfiguration?\n");
+		goto error;
+	}
+
 	memset(db_keys, 0, TPS_NR_KEYS*sizeof(db_key_t));
 	memset(db_vals, 0, TPS_NR_KEYS*sizeof(db_val_t));
 	nr_keys = 0;
@@ -557,6 +784,13 @@ int tps_db_insert_dialog(tps_data_t *td)
 	db_vals[nr_keys].val.str_val = TPS_STRZ(td->s_cseq);
 	nr_keys++;
 
+	if(td->x_context.len>0) {
+		db_keys[nr_keys] = &td_col_x_context;
+		db_vals[nr_keys].type = DB1_STR;
+		db_vals[nr_keys].val.str_val = TPS_STRZ(td->x_context);
+		nr_keys++;
+	}
+
 	if (_tpsdbf.use_table(_tps_db_handle, &td_table_name) < 0) {
 		LM_ERR("failed to perform use table\n");
 		return -1;
@@ -583,6 +817,11 @@ int tps_db_clean_dialogs(void)
 	db_op_t  db_ops[2];
 	int nr_keys;
 
+	if (_tps_db_handle == NULL) {
+		LM_ERR("No database handle - misconfiguration?\n");
+		return -1;
+	}
+
 	nr_keys = 0;
 
 	LM_DBG("cleaning expired dialog records\n");
@@ -629,6 +868,11 @@ int tps_db_insert_branch(tps_data_t *td)
 	db_val_t db_vals[TPS_NR_KEYS];
 	int nr_keys;
 
+	if (_tps_db_handle == NULL) {
+		LM_ERR("No database handle - misconfiguration?\n");
+		goto error;
+	}
+
 	memset(db_keys, 0, TPS_NR_KEYS*sizeof(db_key_t));
 	memset(db_vals, 0, TPS_NR_KEYS*sizeof(db_val_t));
 	nr_keys = 0;
@@ -703,6 +947,43 @@ int tps_db_insert_branch(tps_data_t *td)
 	db_vals[nr_keys].val.str_val = TPS_STRZ(td->s_cseq);
 	nr_keys++;
 
+	db_keys[nr_keys] = &tt_col_a_contact;
+	db_vals[nr_keys].type = DB1_STR;
+	db_vals[nr_keys].val.str_val = TPS_STRZ(td->a_contact);
+	nr_keys++;
+
+	db_keys[nr_keys] = &tt_col_b_contact;
+	db_vals[nr_keys].type = DB1_STR;
+	db_vals[nr_keys].val.str_val = TPS_STRZ(td->b_contact);
+	nr_keys++;
+
+	db_keys[nr_keys] = &tt_col_as_contact;
+	db_vals[nr_keys].type = DB1_STR;
+	db_vals[nr_keys].val.str_val = TPS_STRZ(td->as_contact);
+	nr_keys++;
+
+	db_keys[nr_keys] = &tt_col_bs_contact;
+	db_vals[nr_keys].type = DB1_STR;
+	db_vals[nr_keys].val.str_val = TPS_STRZ(td->bs_contact);
+	nr_keys++;
+
+	db_keys[nr_keys] = &tt_col_a_tag;
+	db_vals[nr_keys].type = DB1_STR;
+	db_vals[nr_keys].val.str_val = TPS_STRZ(td->a_tag);
+	nr_keys++;
+
+	db_keys[nr_keys] = &tt_col_b_tag;
+	db_vals[nr_keys].type = DB1_STR;
+	db_vals[nr_keys].val.str_val = TPS_STRZ(td->b_tag);
+	nr_keys++;
+
+	if(td->x_context.len>0) {
+		db_keys[nr_keys] = &tt_col_x_context;
+		db_vals[nr_keys].type = DB1_STR;
+		db_vals[nr_keys].val.str_val = TPS_STRZ(td->x_context);
+		nr_keys++;
+	}
+
 	if (_tpsdbf.use_table(_tps_db_handle, &tt_table_name) < 0) {
 		LM_ERR("failed to perform use table\n");
 		return -1;
@@ -730,6 +1011,11 @@ int tps_db_clean_branches(void)
 	db_op_t  db_ops[2];
 	int nr_keys;
 
+	if (_tps_db_handle == NULL) {
+		LM_ERR("No database handle - misconfiguration?\n");
+		return -1;
+	}
+
 	nr_keys = 0;
 
 	LM_DBG("cleaning expired branch records\n");
@@ -800,13 +1086,17 @@ int tps_db_clean_branches(void)
 /**
  *
  */
-int tps_db_load_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
+int tps_db_load_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd,
+		uint32_t mode)
 {
-	db_key_t db_keys[4];
-	db_op_t  db_ops[4];
-	db_val_t db_vals[4];
+	db_key_t db_keys[5];
+	db_op_t  db_ops[5];
+	db_val_t db_vals[5];
 	db_key_t db_cols[TPS_NR_KEYS];
 	db1_res_t* db_res = NULL;
+	str sinv = str_init("INVITE");
+	str ssub = str_init("SUBSCRIBE");
+	int bInviteDlg = 1;
 	int nr_keys;
 	int nr_cols;
 	int n;
@@ -818,12 +1108,86 @@ int tps_db_load_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
 	nr_keys = 0;
 	nr_cols = 0;
 
-	db_keys[nr_keys]=&tt_col_x_vbranch;
-	db_ops[nr_keys]=OP_EQ;
-	db_vals[nr_keys].type = DB1_STR;
-	db_vals[nr_keys].nul = 0;
-	db_vals[nr_keys].val.str_val = TPS_STRZ(md->x_vbranch1);
-	nr_keys++;
+	if((get_cseq(msg)->method_id == METHOD_SUBSCRIBE)
+			|| ((get_cseq(msg)->method_id == METHOD_NOTIFY)
+				&& (msg->event && msg->event->len > 0))) {
+		bInviteDlg = 0;
+	}
+
+	if(mode==0) {
+		/* load same transaction using Via branch */
+		db_keys[nr_keys]=&tt_col_x_vbranch;
+		db_ops[nr_keys]=OP_EQ;
+		db_vals[nr_keys].type = DB1_STR;
+		db_vals[nr_keys].nul = 0;
+		db_vals[nr_keys].val.str_val = TPS_STRZ(md->x_vbranch1);
+		nr_keys++;
+	} else {
+		/* load corresponding INVITE transaction using call-id + to-tag */
+		db_keys[nr_keys]=&tt_col_a_callid;
+		db_ops[nr_keys]=OP_EQ;
+		db_vals[nr_keys].type = DB1_STR;
+		db_vals[nr_keys].nul = 0;
+		db_vals[nr_keys].val.str_val = TPS_STRZ(md->a_callid);
+		nr_keys++;
+
+		db_keys[nr_keys]=&tt_col_b_tag;
+		db_ops[nr_keys]=OP_EQ;
+		db_vals[nr_keys].type = DB1_STR;
+		db_vals[nr_keys].nul = 0;
+		db_vals[nr_keys].val.str_val = TPS_STRZ(md->b_tag);
+		nr_keys++;
+
+		db_keys[nr_keys]=&tt_col_s_method;
+		db_ops[nr_keys]=OP_EQ;
+		db_vals[nr_keys].type = DB1_STR;
+		db_vals[nr_keys].nul = 0;
+		db_vals[nr_keys].val.str_val = bInviteDlg ? sinv : ssub;
+		nr_keys++;
+
+		if(md->a_uuid.len>0) {
+			if(md->a_uuid.s[0]=='a') {
+				db_keys[nr_keys]=&tt_col_a_uuid;
+				db_ops[nr_keys]=OP_EQ;
+				db_vals[nr_keys].type = DB1_STR;
+				db_vals[nr_keys].nul = 0;
+				db_vals[nr_keys].val.str_val = TPS_STRZ(md->a_uuid);
+				nr_keys++;
+			} else if(md->a_uuid.s[0]=='b') {
+				db_keys[nr_keys]=&tt_col_b_uuid;
+				db_ops[nr_keys]=OP_EQ;
+				db_vals[nr_keys].type = DB1_STR;
+				db_vals[nr_keys].nul = 0;
+				db_vals[nr_keys].val.str_val = TPS_STRZ(md->a_uuid);
+				nr_keys++;
+			}
+		} else if(md->b_uuid.len>0) {
+			if(md->b_uuid.s[0]=='a') {
+				db_keys[nr_keys]=&tt_col_a_uuid;
+				db_ops[nr_keys]=OP_EQ;
+				db_vals[nr_keys].type = DB1_STR;
+				db_vals[nr_keys].nul = 0;
+				db_vals[nr_keys].val.str_val = TPS_STRZ(md->b_uuid);
+				nr_keys++;
+			} else if(md->b_uuid.s[0]=='b') {
+				db_keys[nr_keys]=&tt_col_b_uuid;
+				db_ops[nr_keys]=OP_EQ;
+				db_vals[nr_keys].type = DB1_STR;
+				db_vals[nr_keys].nul = 0;
+				db_vals[nr_keys].val.str_val = TPS_STRZ(md->b_uuid);
+				nr_keys++;
+			}
+		}
+	}
+
+	if(msg->first_line.type==SIP_REQUEST && md->x_context.len>0) {
+		db_keys[nr_keys]=&tt_col_x_context;
+		db_ops[nr_keys]=OP_EQ;
+		db_vals[nr_keys].type = DB1_STR;
+		db_vals[nr_keys].nul = 0;
+		db_vals[nr_keys].val.str_val = TPS_STRZ(md->x_context);
+		nr_keys++;
+	}
 
 	db_cols[nr_cols++] = &tt_col_rectime;
 	db_cols[nr_cols++] = &tt_col_a_callid;
@@ -839,6 +1203,15 @@ int tps_db_load_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
 	db_cols[nr_cols++] = &tt_col_x_tag;
 	db_cols[nr_cols++] = &tt_col_s_method;
 	db_cols[nr_cols++] = &tt_col_s_cseq;
+	db_cols[nr_cols++] = &tt_col_a_contact;
+	db_cols[nr_cols++] = &tt_col_b_contact;
+	db_cols[nr_cols++] = &tt_col_as_contact;
+	db_cols[nr_cols++] = &tt_col_bs_contact;
+	db_cols[nr_cols++] = &tt_col_a_tag;
+	db_cols[nr_cols++] = &tt_col_b_tag;
+	if(md->x_context.len>0) {
+		db_cols[nr_cols++] = &tt_col_x_context;
+	}
 
 	if (_tpsdbf.use_table(_tps_db_handle, &tt_table_name) < 0) {
 		LM_ERR("failed to perform use table\n");
@@ -852,8 +1225,14 @@ int tps_db_load_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
 	}
 
 	if (RES_ROW_N(db_res) <= 0) {
-		LM_DBG("no stored record for <%.*s>\n",
-				md->x_vbranch1.len, ZSW(md->x_vbranch1.s));
+		if(mode==0) {
+			LM_DBG("no stored record for <%.*s>\n",
+					md->x_vbranch1.len, ZSW(md->x_vbranch1.s));
+		} else {
+			LM_DBG("no stored record for INVITE <%.*s ~ %.*s>\n",
+					md->a_callid.len, ZSW(md->a_callid.s),
+					md->b_tag.len, ZSW(md->b_tag.s));
+		}
 		ret = 1;
 		goto done;
 	}
@@ -875,6 +1254,15 @@ int tps_db_load_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
 	TPS_DATA_APPEND_DB(sd, db_res, n, &sd->x_tag); n++;
 	TPS_DATA_APPEND_DB(sd, db_res, n, &sd->s_method); n++;
 	TPS_DATA_APPEND_DB(sd, db_res, n, &sd->s_cseq); n++;
+	TPS_DATA_APPEND_DB(sd, db_res, n, &sd->a_contact); n++;
+	TPS_DATA_APPEND_DB(sd, db_res, n, &sd->b_contact); n++;
+	TPS_DATA_APPEND_DB(sd, db_res, n, &sd->as_contact); n++;
+	TPS_DATA_APPEND_DB(sd, db_res, n, &sd->bs_contact); n++;
+	TPS_DATA_APPEND_DB(sd, db_res, n, &sd->a_tag); n++;
+	TPS_DATA_APPEND_DB(sd, db_res, n, &sd->b_tag); n++;
+	if(md->x_context.len>0) {
+		TPS_DATA_APPEND_DB(sd, db_res, n, &sd->x_context); n++;
+	}
 
 done:
 	if ((db_res!=NULL) && _tpsdbf.free_result(_tps_db_handle, db_res)<0)
@@ -892,9 +1280,10 @@ error:
 /**
  *
  */
-int tps_storage_load_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
+int tps_storage_load_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd,
+		uint32_t mode)
 {
-	return _tps_storage_api.load_branch(msg, md, sd);
+	return _tps_storage_api.load_branch(msg, md, sd, mode);
 }
 
 /**
@@ -902,9 +1291,9 @@ int tps_storage_load_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
  */
 int tps_db_load_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
 {
-	db_key_t db_keys[4];
-	db_op_t  db_ops[4];
-	db_val_t db_vals[4];
+	db_key_t db_keys[5];
+	db_op_t  db_ops[5];
+	db_val_t db_vals[5];
 	db_key_t db_cols[TPS_NR_KEYS];
 	db1_res_t* db_res = NULL;
 	int nr_keys;
@@ -950,6 +1339,15 @@ int tps_db_load_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
 	}
 	nr_keys++;
 
+	if(md->x_context.len>0) {
+		db_keys[nr_keys]=&td_col_x_context;
+		db_ops[nr_keys]=OP_EQ;
+		db_vals[nr_keys].type = DB1_STR;
+		db_vals[nr_keys].nul = 0;
+		db_vals[nr_keys].val.str_val = TPS_STRZ(md->x_context);
+		nr_keys++;
+	}
+
 	db_cols[nr_cols++] = &td_col_rectime;
 	db_cols[nr_cols++] = &td_col_a_callid;
 	db_cols[nr_cols++] = &td_col_a_uuid;
@@ -971,7 +1369,9 @@ int tps_db_load_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
 	db_cols[nr_cols++] = &td_col_b_srcaddr;
 	db_cols[nr_cols++] = &td_col_s_method;
 	db_cols[nr_cols++] = &td_col_s_cseq;
-
+	if(md->x_context.len>0) {
+		db_cols[nr_cols++] = &td_col_x_context;
+	}
 
 	if (_tpsdbf.use_table(_tps_db_handle, &td_table_name) < 0) {
 		LM_ERR("failed to perform use table\n");
@@ -1015,6 +1415,9 @@ int tps_db_load_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
 	TPS_DATA_APPEND_DB(sd, db_res, n, &sd->b_srcaddr); n++;
 	TPS_DATA_APPEND_DB(sd, db_res, n, &sd->s_method); n++;
 	TPS_DATA_APPEND_DB(sd, db_res, n, &sd->s_cseq); n++;
+	if(md->x_context.len>0) {
+		TPS_DATA_APPEND_DB(sd, db_res, n, &sd->x_context); n++;
+	}
 
 done:
 	if ((db_res!=NULL) && _tpsdbf.free_result(_tps_db_handle, db_res)<0)
@@ -1037,14 +1440,6 @@ int tps_storage_load_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
 	return _tps_storage_api.load_dialog(msg, md, sd);
 }
 
-/**
- *
- */
-int tps_storage_update_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
-{
-	return 0;
-}
-
 #define TPS_DB_ADD_STRV(dcol, dval, cnr, cname, cval) \
 	do { \
 		if(cval.len>0) { \
@@ -1055,15 +1450,129 @@ int tps_storage_update_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
 		} \
 	} while(0)
 
+/**
+ *
+ */
+int tps_db_update_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd,
+		uint32_t mode)
+{
+	db_key_t db_keys[8];
+	db_op_t  db_ops[8];
+	db_val_t db_vals[8];
+	db_key_t db_ucols[TPS_NR_KEYS];
+	db_val_t db_uvals[TPS_NR_KEYS];
+	int nr_keys;
+	int nr_ucols;
+
+	if(_tps_db_handle==NULL)
+		return -1;
+
+	memset(db_ucols, 0, TPS_NR_KEYS*sizeof(db_key_t));
+	memset(db_uvals, 0, TPS_NR_KEYS*sizeof(db_val_t));
+
+	nr_keys = 0;
+	nr_ucols = 0;
+
+	db_keys[nr_keys]=&tt_col_a_uuid;
+	db_ops[nr_keys]=OP_EQ;
+	db_vals[nr_keys].type = DB1_STR;
+	db_vals[nr_keys].nul = 0;
+	if(sd->a_uuid.len>0 && sd->a_uuid.s[0]=='a') {
+		db_vals[nr_keys].val.str_val = TPS_STRZ(sd->a_uuid);
+	} else {
+		if(sd->b_uuid.len<=0) {
+			LM_ERR("no valid dlg uuid\n");
+			return -1;
+		}
+		db_vals[nr_keys].val.str_val = TPS_STRZ(sd->b_uuid);
+	}
+	nr_keys++;
+
+	if(sd->x_context.len>0) {
+		db_keys[nr_keys]=&tt_col_x_context;
+		db_ops[nr_keys]=OP_EQ;
+		db_vals[nr_keys].type = DB1_STR;
+		db_vals[nr_keys].nul = 0;
+		db_vals[nr_keys].val.str_val = TPS_STRZ(sd->x_context);
+		nr_keys++;
+	}
+
+	if(mode & TPS_DBU_CONTACT) {
+		TPS_DB_ADD_STRV(db_ucols, db_uvals, nr_ucols,
+				tt_col_a_contact, md->a_contact);
+		TPS_DB_ADD_STRV(db_ucols, db_uvals, nr_ucols,
+				tt_col_b_contact, md->b_contact);
+	}
+
+	if((mode & TPS_DBU_RPLATTRS) && msg->first_line.type==SIP_REPLY) {
+		if(msg->first_line.u.reply.statuscode>=180
+				&& msg->first_line.u.reply.statuscode<200) {
+
+			db_ucols[nr_ucols] = &tt_col_y_rr;
+			db_uvals[nr_ucols].type = DB1_STR;
+			db_uvals[nr_ucols].val.str_val = TPS_STRZ(md->b_rr);
+			nr_ucols++;
+
+			TPS_DB_ADD_STRV(
+					db_ucols, db_uvals, nr_ucols, tt_col_b_tag, md->b_tag);
+		}
+	}
+	if(nr_ucols==0) {
+		return 0;
+	}
+	if (_tpsdbf.use_table(_tps_db_handle, &tt_table_name) < 0) {
+		LM_ERR("failed to perform use table\n");
+		return -1;
+	}
+
+	if(_tpsdbf.update(_tps_db_handle, db_keys, db_ops, db_vals,
+				db_ucols, db_uvals, nr_keys, nr_ucols)!=0) {
+		LM_ERR("failed to do branch db update for [%.*s]!\n",
+				md->a_uuid.len, md->a_uuid.s);
+		return -1;
+	}
+
+	return 0;
+}
+
+/**
+ *
+ */
+int tps_storage_update_branch(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd,
+		uint32_t mode)
+{
+	int ret;
+
+	if(msg==NULL || md==NULL || sd==NULL)
+		return -1;
+
+	if((md->s_method_id != METHOD_INVITE)
+			&& (md->s_method_id != METHOD_SUBSCRIBE)) {
+		return 0;
+	}
+
+	if(msg->first_line.type==SIP_REPLY) {
+		if(msg->first_line.u.reply.statuscode < 180
+				|| msg->first_line.u.reply.statuscode >= 200) {
+			return 0;
+		}
+	}
+
+	ret = tps_storage_link_msg(msg, md, md->direction);
+	if(ret<0) return -1;
+
+	return _tps_storage_api.update_branch(msg, md, sd, mode);
+}
+
 /**
  *
  */
 int tps_db_update_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd,
 		uint32_t mode)
 {
-	db_key_t db_keys[4];
-	db_op_t  db_ops[4];
-	db_val_t db_vals[4];
+	db_key_t db_keys[8];
+	db_op_t  db_ops[8];
+	db_val_t db_vals[8];
 	db_key_t db_ucols[TPS_NR_KEYS];
 	db_val_t db_uvals[TPS_NR_KEYS];
 	int nr_keys;
@@ -1086,13 +1595,24 @@ int tps_db_update_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd,
 		db_vals[nr_keys].val.str_val = TPS_STRZ(sd->a_uuid);
 	} else {
 		if(sd->b_uuid.len<=0) {
-			LM_ERR("no valid dlg uuid\n");
+			LM_ERR("no valid dlg uuid (%d:%.*s - %d:%.*s)\n",
+					sd->a_uuid.len, sd->a_uuid.len, ZSW(sd->a_uuid.s),
+					sd->b_uuid.len, sd->b_uuid.len, ZSW(sd->b_uuid.s));
 			return -1;
 		}
 		db_vals[nr_keys].val.str_val = TPS_STRZ(sd->b_uuid);
 	}
 	nr_keys++;
 
+	if(sd->x_context.len>0) {
+		db_keys[nr_keys]=&td_col_x_context;
+		db_ops[nr_keys]=OP_EQ;
+		db_vals[nr_keys].type = DB1_STR;
+		db_vals[nr_keys].nul = 0;
+		db_vals[nr_keys].val.str_val = TPS_STRZ(sd->x_context);
+		nr_keys++;
+	}
+
 	if(mode & TPS_DBU_CONTACT) {
 		TPS_DB_ADD_STRV(db_ucols, db_uvals, nr_ucols,
 				td_col_a_contact, md->a_contact);
@@ -1121,6 +1641,40 @@ int tps_db_update_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd,
 			nr_ucols++;
 		}
 	}
+	if(sd->b_tag.len>0 && ((mode & TPS_DBU_BRR) || (mode & TPS_DBU_ARR))) {
+		if(((md->direction == TPS_DIR_DOWNSTREAM)
+					&& (msg->first_line.type==SIP_REPLY))
+				|| ((md->direction == TPS_DIR_UPSTREAM)
+					 && (msg->first_line.type==SIP_REQUEST))) {
+			if(((sd->iflags&TPS_IFLAG_DLGON) == 0) && (mode & TPS_DBU_BRR)) {
+				db_ucols[nr_ucols] = &td_col_b_rr;
+				db_uvals[nr_ucols].type = DB1_STR;
+				db_uvals[nr_ucols].val.str_val = TPS_STRZ(md->b_rr);
+				nr_ucols++;
+			}
+		} else {
+			if(((sd->iflags&TPS_IFLAG_DLGON) == 0) && (mode & TPS_DBU_ARR)) {
+				db_ucols[nr_ucols] = &td_col_a_rr;
+				db_uvals[nr_ucols].type = DB1_STR;
+				db_uvals[nr_ucols].val.str_val = TPS_STRZ(md->a_rr);
+				nr_ucols++;
+				db_ucols[nr_ucols] = &td_col_s_rr;
+				db_uvals[nr_ucols].type = DB1_STR;
+				db_uvals[nr_ucols].val.str_val = TPS_STRZ(md->s_rr);
+				nr_ucols++;
+			}
+		}
+	}
+	if ((mode & TPS_DBU_TIME) && ((sd->b_tag.len > 0)
+			&& ((md->direction == TPS_DIR_UPSTREAM)
+				&& (msg->first_line.type==SIP_REQUEST))
+			&& (msg->first_line.u.request.method_value == METHOD_SUBSCRIBE))) {
+		db_ucols[nr_ucols] = &td_col_rectime;
+		db_uvals[nr_ucols].type = DB1_DATETIME;
+		db_uvals[nr_ucols].val.time_val = time(NULL);
+		nr_ucols++;
+	}
+
 	if(nr_ucols==0) {
 		return 0;
 	}
@@ -1131,7 +1685,7 @@ int tps_db_update_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd,
 
 	if(_tpsdbf.update(_tps_db_handle, db_keys, db_ops, db_vals,
 				db_ucols, db_uvals, nr_keys, nr_ucols)!=0) {
-		LM_ERR("failed to do db update for [%.*s]!\n",
+		LM_ERR("failed to do dialog db update for [%.*s]!\n",
 				md->a_uuid.len, md->a_uuid.s);
 		return -1;
 	}
@@ -1149,12 +1703,14 @@ int tps_storage_update_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd,
 	if(msg==NULL || md==NULL || sd==NULL)
 		return -1;
 
-	if(md->s_method_id != METHOD_INVITE) {
+	if((md->s_method_id != METHOD_INVITE) && (md->s_method_id != METHOD_SUBSCRIBE)) {
 		return 0;
 	}
-	if(msg->first_line.u.reply.statuscode<200
-			|| msg->first_line.u.reply.statuscode>=300) {
-		return 0;
+	if(msg->first_line.type==SIP_REPLY) {
+		if(msg->first_line.u.reply.statuscode < 200
+				|| msg->first_line.u.reply.statuscode >= 300) {
+			return 0;
+		}
 	}
 
 	ret = tps_storage_link_msg(msg, md, md->direction);
@@ -1179,7 +1735,8 @@ int tps_db_end_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
 	if(msg==NULL || md==NULL || sd==NULL || _tps_db_handle==NULL)
 		return -1;
 
-	if(md->s_method_id != METHOD_BYE) {
+	if((md->s_method_id != METHOD_BYE) && !((md->s_method_id == METHOD_SUBSCRIBE)
+				&& (md->expires == 0))) {
 		return 0;
 	}
 
@@ -1204,6 +1761,15 @@ int tps_db_end_dialog(sip_msg_t *msg, tps_data_t *md, tps_data_t *sd)
 	}
 	nr_keys++;
 
+	if(sd->x_context.len>0) {
+		db_keys[nr_keys]=&td_col_x_context;
+		db_ops[nr_keys]=OP_EQ;
+		db_vals[nr_keys].type = DB1_STR;
+		db_vals[nr_keys].nul = 0;
+		db_vals[nr_keys].val.str_val = TPS_STRZ(sd->x_context);
+		nr_keys++;
+	}
+
 	db_ucols[nr_ucols] = &td_col_rectime;
 	db_uvals[nr_ucols].type = DB1_DATETIME;
 	db_uvals[nr_ucols].val.time_val = time(NULL);
